# encoding: utf-8
# license: gpl2p 

### BEGIN LICENSE NOTICE
# This file is part of the interconnect asset ontology framework (IO)
# Copyright (C) 2010 - 2013
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
### END LICENSE NOTICE

### BEGIN AUTHOR LIST
#
### END AUTHOR LIST
 #FIXME add mutex
 module Iof
  module Query
    class Lookup < Ghun::Base::Module

      #NEVER configure dependency cycles
      DEPENDENCIES = {
        :layer1_connections => [:switches, :switch_neighbors],
        :switches => [],
        :switch_neighbors => [:switches],
        :voice_vlans => [],
        :voice_switches => [:voice_vlans],
        :mac_address_to_interface => [],
        :ipv4_address_to_interface => [],
        :ipv4_address_to_network => [:ipv4_address_to_interface,:ipv4_network_to_interface],
        :ipv4_network_to_interface => [],
        :ipv6_address_to_interface => [],
        :ipv6_address_to_network => [:ipv6_address_to_interface,:ipv6_network_to_interface],
        :ipv6_network_to_interface => [],
      }

      def initialize(ontology)
        super(Ghun::Log::Source::QUERY)
        @ontology=ontology
        @timer=Ghun::Timing::Timer.new(Ghun::Log::Source::QUERY)

        @ipv4_addresses=Array.new()
        @ipv6_addresses=Array.new()
        @ipv4_networks=Array.new()
        @ipv6_networks=Array.new()
        @timer.register(:ipv4an,"IPv4 address to network")
        @timer.register(:ipv4ai,"IPv4 address to interface")
        @timer.register(:ipv4ni,"IPv4 network to interface")
        @timer.register(:ipv6an,"IPv6 address to network")
        @timer.register(:ipv6ai,"IPv6 address to interface")
        @timer.register(:ipv6ni,"IPv6 network to interface")
        @timer.register(:ethai,"MAC address to interface")
        @timer.register(:sw,"All switches")
        @timer.register(:neighbors,"All switch neighbors")
        @timer.register(:bfs,"BFS")
        @timer.register(:vvlan,"Voice Vlans")
        @timer.register(:vsw,"Voice switches")
        prepare_reserve_dependencies()
        @lookup_read=0
        @lookup_write=0
        @required_lookup_tables=Array.new()
        @dependency_lookup_tables=Array.new()
        @valid_lookup_tables=Hash.new()
        @invalid_lookup_tables=Array.new()
        @known_query_modules=Hash.new()
      end

      ####
      #### Public lookup table methods
      ####

      def get_lookup_table(table,*args)
        @lookup_read+=1
        if args.nil?() || args.empty?() then
          return @valid_lookup_tables[table]
        else
          result=@valid_lookup_tables[table]
          args.each do |a|
            next if a.nil?()
            if result.is_a?(Hash) && result.include?(a) then              result=result[a]
            else
              result=nil
              break
            end
          end unless result.nil?()
          return result
        end
      end

      def generate_lookup_tables()
        invalid_lookup_tables=@invalid_lookup_tables
        @invalid_lookup_tables=[]
        invalid_lookup_tables.each do |t|
          if @valid_lookup_tables.include?(t) then
            trace "Already updated lookup table '#{t.to_s}' after last request"
            next
          else
            generate_lookup_table(t)
          end
        end
      end

      def register_query_module(mod,required_lookup_tables=[])
        warn "Query module '#{mod.to_s}' already known, updating required lookup tables" if @known_query_modules.include?(mod)
        @known_query_modules[mod]=required_lookup_tables
        generate_required_tables()
        cleanup_lookup_tables()
      end

      def unregister_query_module(mod)
        if @known_query_modules.include?(mod) then
          @known_query_modules.delete(mod)
          generate_required_tables()
          cleanup_lookup_tables()
        else
          warn "Query module '#{mod.to_s}' not known"
        end
      end

      def request_lookup_table_update(table)
        if @valid_lookup_tables.include?(table) then
          r_deps=@reverse_deps[table]
          r_deps.each do |r|
            request_lookup_table_update(r)
          end
          @invalid_lookup_tables << table
          @valid_lookup_tables.delete(table)
        elsif @invalid_lookup_tables.include?(table) then
          trace "Lookup table '#{table.to_s}' already marked for update"
        else
          warn "Lookup table '#{table.to_s}' unknown"
        end
      end

      #FIXME integrate in new structure
      def add_ipv4_address(address)
        @lookup_write+=1
        generate_lookup_table(:ipv4_address_to_network,true,address)
      end

      def add_ipv6_address(address)
        @lookup_write+=1
        generate_lookup_table(:ipv6_address_to_network,true,address)
      end

      ####
      #### Internal lookup table management methods
      ####

    private

      def generate_lookup_table(table,append=false,*args)
        deps=@deps[table]
        deps.each do |d|
          generate_lookup_table(d) unless @valid_lookup_tables.include?(d)
        end unless deps.nil?()
        method="create_lookup_#{table.to_s}".intern
        trace "Running method #{method.to_s}"
        if @valid_lookup_tables.include?(table) && !append then
          trace "Already updated lookup table '#{table.to_s}' after last request"
          return
        end
        result=nil
        begin
          result=self.send(method,*args)
        rescue => e
          error "Failed to generate lookup table '#{table.to_s}'", e
        end

        if append then
          if result.is_a?(Array) then
            @valid_lookup_tables[table]=Array.new() unless @valid_lookup_tables.include?(table) || !@valid_lookup_tables[table].is_a?(Array)
            @valid_lookup_tables[table].merge(result)
          elsif result.is_a?(Hash) then
            @valid_lookup_tables[table]=Hash.new() unless @valid_lookup_tables.include?(table) || !@valid_lookup_tables[table].is_a?(Hash)
            @valid_lookup_tables[table].merge(result)
          else
            warn "No append supported for class '#{result.class}'"
            @valid_lookup_tables[table]=result
          end
        else
          @valid_lookup_tables[table]=result
        end
      end


      def prepare_reserve_dependencies()
        @deps=Query::Lookup::DEPENDENCIES
        @reverse_deps=Hash.new()
        @deps.each do |table,deps|
          deps.each do |d|
            @reverse_deps[d]=Array.new() unless @reverse_deps.include?(d)
            @reverse_deps[d] << table
          end
        end
      end

      def generate_required_tables()
        required_lookup_tables=Array.new()
        dependency_lookup_tables=Array.new()
        @known_query_modules.each do |mod,tables|
          required_lookup_tables.concat(tables)
          tables.each do |t|
            dependency_lookup_tables.concat(@deps[t]) unless @deps[t].nil?()
          end
        end
        required_lookup_tables.uniq!
        dependency_lookup_tables.uniq!
        @required_lookup_tables=required_lookup_tables
        @dependency_lookup_tables=dependency_lookup_tables
      end

      def cleanup_lookup_tables()
        @valid_lookup_tables.each do |table,values|
          @valid_lookup_tables.delete(table) unless @required_lookup_tables.include?(table) || @dependency_lookup_tables.include?(table)
        end
        @invalid_lookup_tables.each do |table,values|
          @invalid_lookup_tables.delete(table) unless @required_lookup_tables.include?(table) || @dependency_lookup_tables.include?(table)
        end
        @required_lookup_tables.each do |t|
          @invalid_lookup_tables << t unless @valid_lookup_tables.include?(t) || @invalid_lookup_tables.include?(t)
        end
      end

      ####
      #### Misc internal methods
      ####

      def output_counter(message=nil)
        debug Ghun::Log::Type::PROFILING, message unless message.nil?()
        @ontology.output_counter()
        debug Ghun::Log::Type::PROFILING, "Handled #{@lookup_read} lookup table reads"
        debug Ghun::Log::Type::PROFILING, "Handled #{@lookup_write} additions to lookup tables"
      end

      def reset_counter()
        @ontology.reset_counter()
        @lookup_read=0
        @lookup_write=0
      end

      def output_and_reset_counter(message=nil)
        output_counter(message)
        reset_counter()
      end

      ####
      ####  Internal lookup table creation
      ####

      def create_lookup_switches(*args)
        switches=Array.new()
        debug "Querying 'switches' data from ontology"
        result=@ontology.query("SELECT DISTINCT ?c WHERE { ?c a no:Switch }")
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        debug "Got no or empty result" if result.nil?() || result.empty?()
        result.each do |r|
          switches << r[:c]
        end
        switches.sort!
        switches.uniq!
        return switches
      end

      def create_lookup_switch_neighbors(*args)
        neighbors=Hash.new()
        get_lookup_table(:switches).each do |node|
          debug "Prefetching neighbors for #{node}"
          neighbors[node.intern]=@ontology.query("SELECT ?c2 ?i1_l1 ?i2_l1 WHERE { ?c2 a no:Switch . no:#{node} no:hasLayer1Connection ?c2 . no:#{node} no:hasLayer1Interface ?i1_l1 . ?c2 no:hasLayer1Interface ?i2_l1 . ?i1_l1 no:isConnectedTo ?i2_l1 }")
          neighbors[node.intern].each do |n|
            n[:c1]=node
          end
        end
        return neighbors
      end

      def create_lookup_voice_vlans(*args)
        debug "Querying 'voice vlans' data from ontology"
        voice_vlans=Array.new()
        result=Array.new()
        result_raw_voice_on_phones=@ontology.query("SELECT DISTINCT ?voice WHERE { ?tmp a no:ManagedNetworkComponent . ?tmp no:isPhone \"true\" . ?tmp no:hasLayer1Interface ?tmp_l1 . ?tmp_l1 no:hasLayer2Interface ?tmp_l2 . ?tmp_l2 no:hasVoiceVlan ?tmp_voice . ?tmp_voice no:hasVlanID ?voice }")
        result_raw_voice_on_phones.map { |v| result << v[:voice].to_i }
        result_raw_voice_on_switches=@ontology.query("SELECT DISTINCT ?voice WHERE { ?tmp a no:Switch . ?tmp no:hasLayer1Interface ?tmp_l1 . ?tmp_l1 no:hasLayer2Interface ?tmp_l2 . ?tmp_l2 no:hasVoiceVlan ?tmp_voice . ?tmp_voice no:hasVlanID ?voice }")
        result_raw_voice_on_switches.map { |v| result << v[:voice].to_i }
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        debug "Got no or empty result" if result.nil?() || result.empty?()
        result.sort!
        result.uniq!
        result.each do |r|
          voice_vlans << r.to_s
        end
        return voice_vlans
      end

      def create_lookup_voice_switches(*args)
        voice_switches=Array.new()
        debug "Querying 'voice switches' data from ontology"
        result=@ontology.query("SELECT DISTINCT ?c WHERE { ?c a no:Switch . ?c no:hasLayer1Connection ?p . ?p a no:ManagedNetworkComponent . ?p no:isPhone \"true\" }")
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        result.each do |s|
          voice_switches << s[:c]
        end

        get_lookup_table(:voice_vlans).each do |vlan|
          result=@ontology.query("SELECT DISTINCT ?c WHERE { ?c a no:Switch . ?c no:hasLayer1Interface ?c_l1 . ?c_l1 no:hasLayer2Interface ?c_l2 . ?c_l2 no:hasVlan ?tmp1 . ?tmp1 no:hasVlanID \"#{vlan}\" }")
          debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
          debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
          result.map { |s| voice_switches << s[:c]}
        end
        voice_switches.sort!
        voice_switches.uniq!
        return voice_switches
      end

      def create_lookup_ipv4_network_to_interface(*args)
        ipv4_network_to_interface=Hash.new
        debug "Querying 'ipv4 network to interface' data from ontology"
        result_mask=@ontology.query("SELECT DISTINCT ?c ?i_l1 ?i_l2 ?i_l3 ?v4_addr ?v4_netmask ?v4_length WHERE {?c a no:Switch .  ?c no:hasLayer1Interface ?i_l1 . ?i_l1 no:hasEthernetInterface ?i_l2 . ?i_l2 no:hasIPv4Interface ?i_l3 . ?i_l3 no:hasIPv4Network ?tmp1 . ?tmp1 no:hasIPv4AddressValue ?v4_addr . ?tmp1 no:hasIPv4NetmaskValue ?v4_netmask . OPTIONAL { ?tmp1 no:hasIPv4NetmaskLength ?v4_length } }")
        result_length=@ontology.query("SELECT DISTINCT ?c ?i_l1 ?i_l2 ?i_l3 ?v4_addr ?v4_netmask ?v4_length WHERE {?c a no:Switch .  ?c no:hasLayer1Interface ?i_l1 . ?i_l1 no:hasEthernetInterface ?i_l2 . ?i_l2 no:hasIPv4Interface ?i_l3 . ?i_l3 no:hasIPv4Network ?tmp1 . ?tmp1 no:hasIPv4AddressValue ?v4_addr . ?tmp1 no:hasIPv4NetmaskLength ?v4_length . OPTIONAL { ?tmp1 no:hasIPv4NetmaskValue ?v4_netmask } }")
        result=(result_mask+result_length).uniq
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        debug "Got no or empty result" if result.nil?() || result.empty?()
        result.each do |r|
          network=nil
          unless r[:v4_length].nil?() then
            debug "Found network #{r[:v4_addr]}/#{r[:v4_length]}"
            @ipv4_networks << IPAddress::IPv4.new("#{r[:v4_addr]}/#{r[:v4_length]}")
            network=Iof::Data.ipv4_network_part(r[:v4_addr],r[:v4_length])
          else
            mask=Iof::Data.ipv4_to_netmask(r[:v4_netmask])
            debug "Found network #{r[:v4_addr]}/#{mask}"
            @ipv4_networks << IPAddress::IPv4.new("#{r[:v4_addr]}/#{mask}")
            network=Iof::Data.ipv4_network_part(r[:v4_addr],r[:v4_netmask])
          end
          interface=Layer3Interface.new(r[:c],r[:i_l1],r[:i_l2],r[:i_l3])
          if ipv4_network_to_interface.include?(network.intern) then
            ipv4_network_to_interface[network.intern] << interface
          else
            ipv4_network_to_interface[network.intern]=[interface]
          end unless network.nil?()
        end
        return ipv4_network_to_interface
      end

      def create_lookup_ipv6_network_to_interface(*args)
        ipv6_network_to_interface=Hash.new()
        debug "Querying 'ipv6 network to interface' data from ontology"
        result=@ontology.query("SELECT DISTINCT ?c ?i_l1 ?i_l2 ?i_l3 ?v6_addr ?v6_length WHERE {?c a no:Switch . ?c no:hasLayer1Interface ?i_l1 . ?i_l1 no:hasEthernetInterface ?i_l2 . ?i_l2 no:hasIPv6Interface ?i_l3 . ?i_l3 no:hasIPv6Network ?tmp1 . ?tmp1 no:hasIPv6AddressValue ?v6_addr . ?tmp1 no:hasIPv6NetmaskLength ?v6_length }")
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        debug "Got no or empty result" if result.nil?() || result.empty?()
        result.each do |r|
          network=Iof::Data.ipv6_network_part(r[:v6_addr],r[:v6_length])
          @ipv6_networks << IPAddress::IPv6.new("#{r[:v6_addr]}/#{r[:v6_length]}")
          interface=Layer3Interface.new(r[:c],r[:i_l1],r[:i_l2],r[:i_l3])
          if ipv6_network_to_interface.include?(network.intern) then
            ipv6_network_to_interface[network.intern] << interface
          else
            ipv6_network_to_interface[network.intern]=[interface]
          end
        end
        return ipv6_network_to_interface
      end

      def create_lookup_ipv4_address_to_network(*args)
        ipv4_address_to_network=Hash.new()
        args.each do |a|
          @ipv4_addresses << IPAddress::IPv4.new(a) unless a.nil?() || a==''
        end unless args.nil?()
        @ipv4_addresses.each do |a|
          debug "Searching network for ipv4 address #{a.to_s}"
          if ipv4_address_to_network.include?(a.to_s.intern) then
            debug "Found existing mapping '#{ipv4_address_to_network[a.to_s.intern].to_s}' for address '#{a.to_s}'" unless ipv4_address_to_network[a.to_s.intern].nil?()
            debug "Found no mapping for address '#{a.to_s}' in prior run" if ipv4_address_to_network[a.to_s.intern].nil?()
            next
          end
          match=nil
          @ipv4_networks.each do |n|
            if match.nil?() then
              match=n
            else
              old_length=match.prefix.to_i
              new_length=n.prefix.to_i
              if old_length < new_length then
                #Better match
                match=n
              elsif old_length > new_length
                #Worse match
                next
              else
                #equal match
                next if match.network.to_s == n.network.to_s
                #conflicting matches
                error "Network '#{match.network.to_s}/#{match.prefix.to_s}' '#{n.network.to_s}/#{n.prefix.to_s}' conflict for address '#{a.to_s}'"
              end
            end if n.include?(a)
          end
          if match.nil?() then
            warn "Found no match for address '#{a.to_s}'"
            ipv4_address_to_network[a.to_s.intern]=nil
          else
            debug "Found match '#{match.network.to_s}' for address '#{a.to_s}'"
            ipv4_address_to_network[a.to_s.intern]=match.network.to_s.intern
          end
        end
        @ipv4_addresses.clear
        return ipv4_address_to_network
      end

      def create_lookup_ipv6_address_to_network(*args)
        ipv6_address_to_network=Hash.new()
        args.each do |a|
          @ipv6_addresses << IPAddress::IPv6.new(a) unless a.nil?() || a==''
        end unless args.nil?()
        @ipv6_addresses.each do |a|
          debug "Searching network for ipv6 address #{a.to_s}"
          if ipv6_address_to_network.include?(a.to_s.intern) then
            debug "Found existing mapping '#{ipv6_address_to_network[a.to_s.intern].to_s}' for address '#{a.to_s}'"
            next
          end
          match=nil
          @ipv6_networks.each do |n|
            if match.nil?() then
              match=n
            else
              old_length=match.prefix.to_i
              new_length=n.prefix.to_i
              if old_length < new_length then
                #Better match
                match=n
              elsif old_length > new_length
                #Worse match
                next
              else
                #equal match
                next if match.network.to_s == n.network.to_s
                #conflicting matches
                error "Network '#{match.network.to_s}/#{match.prefix.to_s}' '#{n.network.to_s}/#{n.prefix.to_s}' conflict for address '#{a.to_s}'"
              end
            end if n.include?(a)
          end
          if match.nil?() then
            warn "Found no match for address '#{a.to_s}'"
            ipv6_address_to_network[a.to_s.intern]=nil
          else
            debug "Found match '#{match.network.to_s}' for address '#{a.to_s}'"
            ipv6_address_to_network[a.to_s.intern]=match.network.to_s.intern
          end
        end
        @ipv6_addresses.clear
        return ipv6_address_to_network
      end

      def create_lookup_layer1_connections_bfs(*args)
        layer1_connections=Hash.new()
        get_lookup_table(:switches).each do |s|
          layer1_connections[s.intern]=bfs(s.intern)
        end
        return layer1_connections
      end
      alias :create_lookup_layer1_connections :create_lookup_layer1_connections_bfs

      def bfs(start_node)
        debug "BFS for start node #{start_node}"
        path=Hash.new()
        visited=Hash.new
        queue=Array.new()
        queue.push([start_node,Iof::Query::Path.new()])
        visited[start_node]=1
        until queue.empty?() do
          node=queue.shift()
          break if node.nil?()
          path[node[0]]=node[1]
          neighbors=get_lookup_table(:switch_neighbors,node[0])
          neighbors.each do |n|
            if visited[n[:c2].intern].nil?() then
              visited[n[:c2].intern]=1
              p=node[1].clone()
              p << Iof::Query::PathEntry.new(n[:c1],n[:i1_l1])
              p << Iof::Query::PathEntry.new(n[:c2],n[:i2_l1])
              queue.push([n[:c2].intern,p])
            end
          end
        end
        path
      rescue =>e
        error "Failed to calculate path for start node #{start_node}", e
      end

      def create_lookup_ipv4_address_to_interface(*args)
        ipv4_address_to_interface=Hash.new()
        debug "Querying 'ipv4 address to interface' data from ontology"
        result=@ontology.query("SELECT DISTINCT ?c ?i_l1 ?i_l2 ?i_l3 ?v4 WHERE {?c a no:ManagedNetworkComponent . ?c no:hasLayer1Interface ?i_l1 . ?i_l1 no:hasEthernetInterface ?i_l2 . ?i_l2 no:hasIPv4Interface ?i_l3 . ?i_l3 no:hasIPv4Address ?tmp1 . ?tmp1 no:hasIPv4AddressValue ?v4}")
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        debug "Got no or empty result" if result.nil?() || result.empty?()
        result.each do |r|
          v4=r[:v4]
          @ipv4_addresses << IPAddress::IPv4.new(v4)
          interface=Layer3Interface.new(r[:c],r[:i_l1],r[:i_l2],r[:i_l3])
          if ipv4_address_to_interface.include?(v4.intern) then
            ipv4_address_to_interface[v4.intern] << interface
            warn "IPv4 address #{v4} is not unique", "Got '#{ipv4_address_to_interface[v4.intern].inspect}' and '#{interface.inspect}'"
          else
            ipv4_address_to_interface[v4.intern]=[interface]
          end
        end
        return ipv4_address_to_interface
      end

      def create_lookup_ipv6_address_to_interface(*args)
        ipv6_address_to_interface=Hash.new()
        debug "Querying 'ipv6 address to interface' data from ontology"
        result=@ontology.query("SELECT DISTINCT ?c ?i_l1 ?i_l2 ?i_l3 ?v6 WHERE {?c a no:ManagedNetworkComponent . ?c no:hasLayer1Interface ?i_l1 . ?i_l1 no:hasEthernetInterface ?i_l2 . ?i_l2 no:hasIPv6Interface ?i_l3 . ?i_l3 no:hasIPv6Address ?tmp1 . ?tmp1 no:hasIPv6AddressValue ?v6}")
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        debug "Got no or empty result" if result.nil?() || result.empty?()
        result.each do |r|
          v6=r[:v6]
          @ipv6_addresses << IPAddress::IPv6.new(v6)
          interface=Layer3Interface.new(r[:c],r[:i_l1],r[:i_l2],r[:i_l3])
          if ipv6_address_to_interface.include?(v6.intern) then
            ipv6_address_to_interface[v6.intern] << interface
            warn "IPv6 address #{v6} is not unique", "Got '#{ipv6_address_to_interface[v6.intern].inspect}' and '#{interface.inspect}'"
          else
            ipv6_address_to_interface[v6.intern]=[interface]
          end
        end
        return ipv6_address_to_interface
      end

      def create_lookup_mac_address_to_interface(*args)
        mac_address_to_interface=Hash.new()
        debug "Querying 'mac address to interface' data from ontology"
        result=@ontology.query("SELECT DISTINCT ?c ?i_l1 ?i_l2 ?m WHERE {?c a no:ManagedNetworkComponent . ?c no:hasLayer1Interface ?i_l1 . ?i_l1 no:hasEthernetInterface ?i_l2 . ?i_l2 no:hasMacAddress ?tmp1 . ?tmp1 no:hasMacAddressValue ?m}")
        debug "Got result #{result.inspect}" unless result.nil?() || result.empty?()
        debug "Got no or empty result" if result.nil?() || result.empty?()
        result.each do |r|
          mac=r[:m]
          interface=Layer2Interface.new(r[:c],r[:i_l1],r[:i_l2])
          if mac_address_to_interface.include?(mac.intern) then
            next if mac_address_to_interface[mac.intern]==interface
            warn "MAC address #{mac} is not unique, using only first occurence", "Got '#{mac_address_to_interface[mac.intern].inspect}' and '#{interface.inspect}'"
          else
            mac_address_to_interface[mac.intern]=interface
          end
        end
        return mac_address_to_interface
      end
    end
  end
end

#FIXME old code

#      # Floyd warshall dijkstra may be needed again if connections become weighted
#      def create_lookup_layer1_connections_floyd_warshall()
#        debug "Querying 'layer1 connections (floyd warshall)' data from ontology"
#        max_fixnum=2**((0.size*8)-2)-1
#        weight=Hash.new()
#        @layer1_connections=Hash.new()
#        @switches.each do |src|
#          debug "Creating weight matrix line for src #{src}"
#          weight[src.intern]=Hash.new()
#          @layer1_connections[src.intern]=Hash.new()
#          @switches.each do |dest|
#            if src==dest then
#              weight[src.intern][dest.intern]=0
#              @layer1_connections[src.intern][dest.intern]=Iof::Query::Path.new()
#            else
#              weight[src.intern][dest.intern]=max_fixnum
#              @layer1_connections[src.intern][dest.intern]=nil
#            end
#          end
#          @neighbors[src.intern].each do |n|
#            weight[src.intern][n[:c2].intern]=1
#            @layer1_connections[src.intern][n[:c2].intern]=Iof::Query::Path.new
#            @layer1_connections[src.intern][n[:c2].intern] << Iof::Query::PathEntry.new(src,n[:i1_l1])
#            @layer1_connections[src.intern][n[:c2].intern] << Iof::Query::PathEntry.new(n[:c2],n[:i2_l1])
#          end
#        end
#
#        @switches.each do |dest| #k
#          debug "Using destination #{dest}"
#          @switches.each do |src| #i
#            #Optimization: if no path from i to k, no need to test k to j
#            @switches.each do |s| #j
#              dist=weight[src.intern][dest.intern] + weight[dest.intern][s.intern]
#              if dist < weight[src.intern][s.intern] then
#                weight[src.intern][s.intern]=dist
#                @layer1_connections[src.intern][s.intern]=@layer1_connections[src.intern][dest.intern].clone
#                @layer1_connections[src.intern][s.intern] << @layer1_connections[dest.intern][s.intern]
#              end
#            end if weight[src.intern][dest.intern] < max_fixnum
#          end
#        end
#      end
#
#      # TODO: replace distance hash with priority queue
#      def create_lookup_layer1_connections_dijkstra()
#        debug "Querying 'layer1 connections (dijkstra)' data from ontology"
#        distance=Hash.new()
#        predecessor=Hash.new
#        all_nodes=Array.new
#        debug "Initalizing needed data structures for dijkstra algorithm"
#        dijkstra_init(@switches,nil,distance,predecessor,all_nodes)
#        @switches.each do |s_src|
#          debug "Searching paths to all nodes from #{s_src}"
#          dijkstra_result=dijkstra(@switches,s_src.intern,distance.clone,predecessor.clone,all_nodes.clone)
#          debug "Calculated paths to all nodes from #{s_src}"
#          @switches.each do |s_dest|
#            path=nil
#            path=dijkstra_single_path(s_dest.intern, dijkstra_result) unless dijkstra_result.nil?()
#            path=Iof::Query::Path.new() if s_src==s_dest
#            @layer1_connections[s_src.intern]=Hash.new unless @layer1_connections.include?(s_src.intern)
#            @layer1_connections[s_src.intern][s_dest.intern]=path
#          end unless @switches.nil?
#        end unless @switches.nil?
#      end
#
#      def dijkstra(graph,source,distance,predecessor,all_nodes)
#        max_fixnum=2**((0.size*8)-2)-1
#        dijkstra_init(graph, source, distance, predecessor, all_nodes)
#        until all_nodes.empty?()
#          min_distance=max_fixnum
#          next_node=nil
#          all_nodes.each do |n|
#            this_distance=distance[n]
#            if this_distance < min_distance then
#              next_node=n
#              min_distance=this_distance
#            end
#          end
#          if next_node.nil?() then
#            debug "Terminating dijkstra, all remaining nodes are unreachable", "Remaining nodes #{all_nodes.inspect}"
#            break
#          end
#          all_nodes.delete(next_node)
#          neighbors=@neighbors[next_node]
#          neighbors.each do |r|
#            neighbor_node=r[:c2].intern
#            alternative_length = distance[next_node] + 1
#            if alternative_length < distance[neighbor_node] then
#              distance[neighbor_node]=alternative_length
#              predecessor[neighbor_node]=r
#            end
#          end
#        end
#        return predecessor
#      end
#
#      def dijkstra_init(graph,source,distance,predecessor,all_nodes)
#        if source.nil?()
#          max_fixnum=2**((0.size*8)-2)-1
#          graph.each do |node|
#            name=node.intern
#            all_nodes << name
#            distance[name]=max_fixnum
#            predecessor[name]=nil
#          end
#        else
#          distance[source]=0
#        end
#      end
#
#      def dijkstra_single_path(destination,predecessor)
#        path=Iof::Query::Path.new
#        pre=predecessor[destination]
#        until pre.nil?()
#          path << Iof::Query::PathEntry.new(pre[:c2],pre[:i2_l1])
#          path << Iof::Query::PathEntry.new(pre[:c1],pre[:i1_l1])
#          pre=predecessor[pre[:c1].intern]
#        end
#        if path.unchanged? then
#          return nil
#        else
#          path.reverse!
#          return path
#        end
#      end
