# encoding: utf-8
# license: gpl2p 

### BEGIN LICENSE NOTICE
# This file is part of the interconnect asset ontology framework (IO)
# Copyright (C) 2010 - 2013
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
### END LICENSE NOTICE

### BEGIN AUTHOR LIST
#
### END AUTHOR LIST

module Ghun
  module Base
    class Connection  < Ghun::Base::Module
      def initialize(mode,type=:unknown,log_source=Ghun::Log::Source::BASE,log_type=Ghun::Log::Type::BASE)
        super(log_source,log_type)
        @async_mode=(mode==:async || mode==:combined)
        @sync_mode=(mode==:sync || mode==:combined)
        @combined_mode=(mode==:misc)
        if @async_mode then
          @shutdown=false
          @requests=::Queue.new()
          @responses=::Queue.new()
          @response_mutex=Mutex.new()
          @thread=Ghun::Base::Thread.new("#{self.class}ExecutionRunner#{self.__id__}")
          @thread.start() {
            execution_runner()
          }
        end
        @execution_mutex=Monitor.new()
        @sync_commands_pending=0
        @pending_mutex=Mutex.new()
        @type=type
        disable_retries()
      end

      def configure_retries(execute_retries,execute_retry_wait,execute_retry_wait_first_time)
        @retry_mode=true
        @execute_retries=execute_retries
        @execute_retry_wait=execute_retry_wait
        @execute_retry_wait_first_time=execute_retry_wait_first_time
      end

      def disable_retries()
        @working_before=false
        @retry_mode=false
        @execute_retries=0
        @execute_retry_wait=0
        @execute_retry_wait_first_time=0
      end

      def enable_wait_reporting(thread_handler)
        @thread_handler=thread_handler
      end

      def disable_wait_reporting()
        @thread_handler=nil
      end

      def execute_request(args)
        raise NotSupportedInAsyncModeError unless @combined_mode || @sync_mode
        response=nil
        @pending_mutex.synchronize {
          @sync_commands_pending+=1
        }
        @execution_mutex.synchronize {
          begin
            ensure_connection() if self.respond_to?(:ensure_connection,true)
            if args.is_a?(Array)
              response=Array.new()
              args.each do |argument|
                response << execute_retry_helper(argument)
              end
            else
              response=execute_retry_helper(args)
            end
          rescue => e
            error "Failed to execute '#{args.inspect}' for connection type #{self.class}", e
            raise
          ensure
            disconnect() if self.respond_to?(:disconnect,true) && !@keep_connection
          end
        }
        @pending_mutex.synchronize {
          @sync_commands_pending-=1
        }
        return response
      end

      def enqueue_request(id,args)
        raise NotSupportedInSyncModeError unless @combined_mode || @async_mode
        return nil if @shutdown
        @requests.enq({:id => id, :arguments => args})
      end

      def get_response()
        raise NotSupportedInSyncModeError unless @async_mode
        @response_mutex.synchronize {
          return @responses.pop()
        }
      end

      def get_response_non_blocking()
        raise NotSupportedInSyncModeError unless @async_mode
        @response_mutex.synchronize {
          return nil if @responses.length()==0
          return @responses.pop()
        }
      end

      def stop_connection()
        raise NotSupportedInSyncModeError unless @async_mode
        @shutdown=true
        @response_mutex.synchronize {
          @responses.enq(nil)
        }
      end

    private

      def reported_sleep(duration)
        @thread_handler.report_thread_sleep(duration, @type) unless @thread_handler.nil?()
        sleep duration
      ensure
        @thread_handler.report_thread_wakeup(duration, @type) unless @thread_handler.nil?()
      end

      def reported_execution(&block)
        @thread_handler.report_thread_longexec_start(@type) unless @thread_handler.nil?()
        block.call
      ensure
        @thread_handler.report_thread_longexec_stop(@type) unless @thread_handler.nil?()
      end

      def execution_runner()
        return nil unless @async_mode
        until @shutdown do
          begin
            request=@requests.pop()
            if request.nil?()
              sleep(0.1)
              next
            end
            ensure_connection() if self.respond_to?(:ensure_connection,true)
            response=nil
            if request[:args].is_a?(Array)
              response=Array.new()
              request[:args].each do |req|
                wait=true
                while wait
                  @pending_mutex.synchronize {
                    wait&&=(@sync_commands_pending>=0)
                  }
                  sleep 0.1
                end
                @execution_mutex.synchronize {
                  response << execute_retry_helper(req)
                }
              end
            else
              wait=true
              while wait
                @pending_mutex.synchronize {
                  wait&&=(@sync_commands_pending>=0)
                }
                sleep 0.1
              end
              @execution_mutex.synchronize {
                response=execute_retry_helper(request[:args])
              }
            end
            disconnect() if self.respond_to?(:disconnect,true) && !@keep_connection
            @response_mutex.synchronize {
              @responses.enq({:id => request[:id], :response => response})
            }
          rescue => e
            error "Failed to execute '#{request.inspect}' for connection type #{self.class}", e
          ensure
            disconnect() if self.respond_to?(:disconnect,true) && !@keep_connection
          end
        end
      end

      def execute_retry_helper(args)
        if @retry_mode
          count=0
          response=nil
          while @retry_mode && count<=@execute_retries && response.nil?()
            if count > 0
              disconnect() if self.respond_to?(:disconnect,true)
              if @working_before
                reported_sleep(@execute_retry_wait)
              else
                reported_sleep(@execute_retry_wait_first_time)
              end if count>0
              connect() if self.respond_to?(:connect,true)
            end
            begin
              count+=1
              response=execute_helper(args)
            rescue => e
              warn "Failed execution try", e
              response=nil
            end
          end
          if response.nil?()
            raise RetryFailedError, "Failed to execute #{args.inspect} after #{@execute_retries} retries"
          else
            @working_before=true
            return response
          end
        else
          return execute_helper(args)
        end
      end

      def execute_helper(args={})
        raise AbstractClassError, "Reimplement 'execute_helper' to make use of connection type '#{self.class}'"
      end
    end

    class StatefullConnection < Connection
      def initialize(mode,type=:statefull,keep_connection=true,log_source=Ghun::Log::Source::BASE,log_type=Ghun::Log::Type::BASE)
        super(mode,type,log_source,log_type)
        @keep_connection=keep_connection
        @connecting_mutex=Monitor.new()
        disable_retries()
      end

      def configure_retries(connect_retries,connect_retry_wait,connect_retry_wait_first_time,execute_retries,execute_retry_wait,execute_retry_wait_first_time)
        super(execute_retries,execute_retry_wait,execute_retry_wait_first_time)
        @connect_retries=connect_retries
        @connect_retry_wait=connect_retry_wait
        @connect_retry_wait_first_time=connect_retry_wait_first_time
      end

      def disable_retries()
        super()
        @connect_retries=0
        @connect_retry_wait=0
        @connect_retry_wait_first_time=0
      end

      def keep_connection?()
        @keep_connection
      end

      def keep_connection=(keep_connection)
        @keep_connection=keep_connection
        disconnect() if keep_connection && connected?()
      end

      def connect()
        @connecting_mutex.synchronize {
          @connected=connect_retry_helper()
        }
      end

      def disconnect()
        @connecting_mutex.synchronize {
          @connected=disconnect_helper()
        }
      end

      def connected?()
        @connecting_mutex.synchronize {
          return @connected
        }
      end

    private
      def ensure_connection()
        connect() unless connected?()
        raise Ghun::Base::ConnectionError,"Failed to create connection of class #{self.class}" unless connected?()
      end

      def connect_retry_helper()
        if @retry_mode
          count=0
          response=nil
          while @retry_mode && count<=@connect_retries && response.nil?()
            if @working_before
              reported_sleep(@connect_retry_wait)
            else
              reported_sleep(@connect_retry_wait_first_time)
            end if count>0
            begin
              count+=1
              reported_execution() { response=connect_helper() }
            rescue => e
              warn "Failed connection try", e
              response=nil
            end
          end
          unless connected?()
            raise RetryFailedError, "Failed to connect after #{@connect_retries} retries"
          else
            @working_before=true
            return response
          end
        else
          return connect_helper()
        end
      end


      def connect_helper()
        raise AbstractClassError, "Reimplement 'connect_helper' to make use of connection type '#{self.class}'"
      end

      def disconnect_helper()
        raise AbstractClassError, "Reimplement 'disconnect_helper' to make use of connection type '#{self.class}'"
      end
    end

    class StatelessConnection < Connection
      def initialize(mode,type=:stateless,log_source=Ghun::Log::Source::BASE,log_type=Ghun::Log::Type::BASE)
        super(mode,type,log_source,log_type)
      end
    end

    class ConnectionHandler < Ghun::Base::SimpleHandler
      def initialize(log_source=Ghun::Log::Source::BASE,log_type=Ghun::Log::Type::BASE)
        super(log_source,log_type)
      end
    end
  end
end
