# encoding: utf-8
# license: gpl2p 

### BEGIN LICENSE NOTICE
# This file is part of the interconnect asset ontology framework (IO)
# Copyright (C) 2010 - 2013
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
### END LICENSE NOTICE

### BEGIN AUTHOR LIST
#
### END AUTHOR LIST

require 'thread'
require 'timeout'

module Ghun
  module Base
    class Thread
      include Ghun::Base::Logging
      def initialize(name,log_source=Ghun::Log::Source::BASE,log_type=Ghun::Log::Type::BASE)
        init_logging(log_source,log_type,Log::Level::DEBUG)
        @name=name
        @state=:stopped
        @state_mutex=Mutex.new()
        @thread=nil
      end

      def self.run(name,timeout=0,log_source=Ghun::Log::Source::BASE,log_type=Ghun::Log::Type::BASE,&block)
        thread=self.new(name,log_source,log_type)
        thread.start(timeout,&block)
        return thread
      end

      def run()
        raise AbstractClassError, "run needs to be reimplemented"
      end
      private :run

      def start(timeout=0,&block)
        status()
        @state_mutex.lock()
        if @state==:running then
          warn "Thread #{@name} is running already"
        else
          @state=:"starting"
          if timeout==0 then
            trace "Starting thread #{@name} without timeout"
            @thread=::Thread.new() do
              begin
                if block_given?() then
                  block.call
                else
                  self.run()
                end
              rescue => e
                error "Thread #{@name} terminated with error", e
              else
                trace "Thread #{@name} terminated"
              end
            end
            @state=:"running"
          else
            trace "Starting thread #{@name} with #{timeout} seconds timeout"
            begin
              Timeout.timeout(timeout) {
                begin
                  if block_given?() then
                    block.call
                  else
                    self.run()
                  end
                rescue => e
                  error "Thread #{@name} terminated with error", e
                else
                  trace "Thread #{@name} terminated"
                end
              }
            rescue Timeout::Error => e
              warn "Thread #{@name} timed out"
            rescue => e
              error "Thread #{@name} terminated with error", e
            else
              trace "Thread #{@name} terminated"
            end
            @state=:"running"
          end
        end
      rescue => e
        error "Failed to start thread '#{@name}' of type '#{self.class}'", e
        @state=:stopped
      ensure
        @state_mutex.unlock()
      end

      def exit
        !@thread.nil?() && @thread.exit()
        until !@thread.status do
          sleep 0.01
        end unless @thread.nil?()
        status()
      end
      alias :kill :exit
      alias :terminate :exit

      def join()
        @thread.nil?() && @thread.join()
        until !@thread.status do
          sleep 0.01
        end unless @thread.nil?()
        status()
      end

      def status()
        @state_mutex.lock()
        if @thread.nil?() then
          @state=:stopped
        elsif !@thread.status then
          @state=:stopped
          @thread=nil
        end
        @state_mutex.unlock()
        !@thread.nil?() && @thread.status()
      end

      def alive?()
        !@thread.nil?() && @thread.alive?()
      end

      def stop?()
        !@thread.nil?() && @thread.stop?()
      end
    end
  end
end
