# encoding: utf-8
# license: gpl2p 

### BEGIN LICENSE NOTICE
# This file is part of the interconnect asset ontology framework (IO)
# Copyright (C) 2010 - 2013
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
### END LICENSE NOTICE

### BEGIN AUTHOR LIST
#
### END AUTHOR LIST
module Ghun
  module Log
    class Handler
      include Ghun::Base::Configuration
      Ghun::Base::Blackboard.config.declare_key(:"log.developer_mode",:bool,false)
      Ghun::Base::Blackboard.config.declare_key(:"log.threaded",:bool,false)
      class Queues < Ghun::Base::AutoloadingHandler
        Ghun::Base::Blackboard.config.declare_key(:"log.queue",:array,[:"default",:"development"])
        def initialize()
          super(Log::Source::LOG)
          reload_config
          @shutdown=false
        end

        def member_class(type)
          Log::Queue
        end

        def init_member_list
          @member_list=Array.new()
          member_list=config[:"log.queue"]
          member_list.each do |m|
            config.declare_dynamic_key("log.queue.#{m.to_s}.enabled".intern,:bool,true)
            if m==:development
              config.declare_dynamic_key("log.queue.#{m.to_s}.only_developer".intern,:bool,true)
              config.declare_dynamic_key("log.queue.#{m.to_s}.loglevel".intern,:log_level,Log::Level::DEBUG)
              config.declare_dynamic_key("log.queue.#{m.to_s}.storage".intern,:array,["file"])
            else
              config.declare_dynamic_key("log.queue.#{m.to_s}.only_developer".intern,:bool,false)
              config.declare_dynamic_key("log.queue.#{m.to_s}.loglevel".intern,:log_level,Log::Level::WARNING)
              config.declare_dynamic_key("log.queue.#{m.to_s}.storage".intern,:array,["stderr","file"])
            end
            config.declare_dynamic_key("log.queue.#{m.to_s}.identifier".intern,:string,m.to_s)
            config.declare_dynamic_key("log.queue.#{m.to_s}.logsource".intern,:log_source,Log::Source::ALL)
            config.declare_dynamic_key("log.queue.#{m.to_s}.logtype".intern,:log_type,Log::Type::ALL)
            @member_list << m.intern
          end unless member_list.nil?()
        end
      end

      STORAGE_SELECTOR=Ghun::Base::Selector.new("Logging",[:storage_type],false,Ghun::Log::Source::LOG,Ghun::Log::Type::BASE)
      class Storages < Ghun::Base::AutoloadingHandler
        def initialize()
          super(Log::Source::LOG)
          reload_config
        end

        def member_class(type)
          result=STORAGE_SELECTOR.select_member(type)
          if result.nil?()
            self.error "Unknown log storage engine #{type.to_s}"
            raise InvalidStorageEngineError, "Unknown config storage engine #{type}"
          end
          return result
        end

        def init_member_list
          @member_list=Array.new()
          member_list=Array.new()
          STORAGE_SELECTOR.members.values.each do |m|
            member_list << m[:values][0][0]
          end unless STORAGE_SELECTOR.members.nil?()
          member_list.each do |m|
            config.declare_dynamic_key("log.storage.#{m.to_s}.enabled".intern,:bool,false)
            if m==:file
              config.declare_dynamic_key("log.storage.#{m.to_s}.loglevel".intern,:log_level,Log::Level::DEBUG)
            else
              config.declare_dynamic_key("log.storage.#{m.to_s}.loglevel".intern,:log_level,Log::Level::WARNING)
            end
            config.declare_dynamic_key("log.storage.#{m.to_s}.logsource".intern,:log_source,Log::Source::ALL)
            config.declare_dynamic_key("log.storage.#{m.to_s}.logtype".intern,:log_type,Log::Type::ALL)
            @member_list << m.intern
          end unless member_list.nil?()
        end
      end

      def initialize
        register
        enable_buffering
        @queues_mutex=Monitor.new()
        @messages=::Queue.new()
        @storages=Storages.new()
        @queues=Queues.new()
        reload_config(true)
        @storages.enable_all
        @queues.enable_all
        disable_buffering
      end
      attr_reader :storages,:queues

      def reload_config(first=false)
        enable_buffering unless first
        self.debug(Log::Source::LOG,Log::Type::BASE,"Reloading log configuration")
        @develop=config[:"log.developer_mode"]
        @threaded=config[:"log.threaded"]
        disable_buffering unless first
      end

      def enable_buffering()
        @buffering=true
      end

      def disable_buffering()
        @queues_mutex.synchronize {
          message=@messages.pop()
          @queues.members.each do |name,queue|
            queue.log(message) if !queue.nil?() && queue.enabled?
          end unless @queues.members.nil?()
        }
        @buffering=false
      end

      def is_threaded?()
        return @threaded
      end

      def get_storage(name)
        @storages[name]
      end
      alias []  get_storage

      def log(loglevel,logsource,logtype,messages)
        if messages.is_a? Enumerable then
          messages.each do |message|
            log loglevel,logsource,logtype,message
          end
        else
          if messages.is_a? Exception then
            log loglevel,logsource,logtype,"Exception: #{messages.class}"
            log loglevel,logsource,logtype,messages.message
            log loglevel,logsource,logtype,messages.backtrace
          elsif (@buffering || @threaded) && messages.is_a?(String) then
            @messages.enq(Log::Message.new(loglevel,logsource,logtype,messages))
          elsif !@buffering && !@threaded && messages.is_a?(String) then
            @queues_mutex.synchronize {
              @queues.members.each do |name,queue|
                queue.log(Log::Message.new(loglevel,logsource,logtype,messages)) if !queue.nil?() && queue.enabled?
              end unless @queues.members.nil?()
            }
          elsif messages.respond_to? :to_s then
            log loglevel,logsource,logtype,"Class: #{messages.class}; Content: #{messages.to_s}"
          else
            log loglevel,logsource,logtype,"Class: #{messages.class}; content not accessible"
          end
        end
      end

      def log_all_messages
        return nil unless @threaded
        locked=@queues_mutex.try_lock()
        unless locked then
          log(Log::Level::WARNING,Log::Source::LOG,Log::Type::BASE,"Logger already running")
          return
        end
        begin
          @shutdown=false
          until @shutdown do
            message=@messages.pop()
            @queues.members.each do |name,queue|
              queue.log(message) if !queue.nil?() && queue.enabled?
            end unless @queues.members.nil?()
          end
        rescue => e
          log(Log::Level::ERROR,Log::Source::LOG,Log::Type::BASE,["Error while logging messages",e])
        ensure
          @queues_mutex.unlock()
        end
      end

      def stop_log_all_messages
        return nil unless @threaded
        @shutdown=true
        log(Log::Level::DEBUG,Log::Source::LOG,Log::Type::BASE,"Stopped message logging")
      end

      def log_all_messages_non_blocking()
        return nil unless @threaded
        locked=@queues_mutex.try_lock()
        unless locked then
          log(Log::Level::WARNING,Log::Source::LOG,Log::Type::BASE,"Logger already running")
          return
        end
        begin
          @shutdown=false
          until @shutdown || @messages.length()==0 do
            message=@messages.pop()
            @queues.members.each do |name,queue|
              queue.log(message) if !queue.nil?() && queue.enabled?
            end unless @queues.members.nil?()
          end
        rescue => e
          log(Log::Level::ERROR,Log::Source::LOG,Log::Type::BASE,["Error while logging messages",e])
        ensure
          @queues_mutex.unlock()
        end
      end


      def debug(logsource,logtype,*messages)
        log Level::DEBUG,logsource,logtype,messages
      end

      def trace(logsource,logtype,*messages)
        log Level::DEBUG,logsource,logtype,messages
      end

      def info(logsource,logtype,*messages)
        log Level::INFO,logsource,logtype,messages
      end

      def warn(logsource,logtype,*messages)
        log Level::WARNING,logsource,logtype,messages
      end

      def error(logsource,logtype,*messages)
        log Level::ERROR,logsource,logtype,messages
      end

      def fatal(logsource,logtype,*messages)
        log Level::FATAL,logsource,logtype,messages
      end
    end
  end
end

