# encoding: utf-8
# license: gpl3p

### BEGIN LICENSE NOTICE
# This file is part of the interconnect asset ontology framework (IO)
# Copyright (C) 2010 - 2013
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE NOTICE

### BEGIN AUTHOR LIST
#
### END AUTHOR LIST
require "socket"
require "openssl"
module Iof
  module Control
    class ServerBase < Ghun::Base::Module
      include Ghun::Base::Configuration
      def initialize()
        super(Ghun::Log::Source::CONTROL,Ghun::Log::Source::BASE)
        init_configuration()
        @connection=Array.new()
      end

      def listen(socket)
        loop do
          begin
            connection=socket.accept
            @connection << connection
            info "Client '#{connection.peeraddr[3]}:#{connection.peeraddr[1]}' connected"
            Ghun::Base::Thread.run("Client '#{connection.peeraddr[3]}:#{connection.peeraddr[1]}'",0,Ghun::Log::Source::CONTROL) do
              handle_connection(connection)
            end
          rescue => e
            error "Accepting connection failed",e
          end
        end
      end

      def handle_connection(connection)
        disconnect=false
        while !disconnect && (line_in=connection.gets())
          response=nil
          begin
            request=decode(line_in)
            debug Ghun::Log::Type::RECEIVE, "Got message #{request.inspect}"
            response=handle_request(request,connection)
            disconnect=true if response.is_a?(FalseClass)
            raise Iof::Control::UncleanGenericError, "Got empty response after command execution" if response.nil?()
          rescue => e
            error "Could not create response for client '#{connection.peeraddr[3]}:#{connection.peeraddr[1]}'", e
            response=VisaResponse.translate_from_exception(e)
          end
          unless disconnect
            line_out=encode(response)
            connection.puts(line_out)
          end
        end
        if disconnect
          peer="#{connection.peeraddr[3]}:#{connection.peeraddr[1]}"
          connection.close()
          info "Client '#{peer}' disconnected"
        end
      rescue => e
        error "Connection to '#{connection.peeraddr[3]}:#{connection.peeraddr[1]}' terminated with an exception",e
      end

      def decode(line_in)
        request=line_in.chomp
        begin
          request=Base64.strict_decode64(request)
        rescue => e
          error "Decoding Base64 failed", e
          raise Iof::Control::UnexpectedTransferEncodingError, "BASE64 encoding is corrupted or unknown encoding is used"
        end
        request.force_encoding("UTF-8")
        begin
          request=YAML.load(request)
        rescue => e
          error "Decoding YAML failed", e
          raise Iof::Control::UnexpectedMessageType, "YAML encoding is corrupted or message type is unknown"
        end
        return request
      end

      def encode(response)
        line_out=response.to_yaml()
        line_out=Base64.strict_encode64(line_out)
        return line_out
      rescue => e
        error "Failed to encode message #{response.inspect()}", e
        #returning nil in case of an error would trigger the regular error case handling,
        #but may cause a endless loop, when failing during encoding an error message
        return ""
      end

      def handle_request(request,connection=nil)
        raise Ghun::Base::AbstractClassError, "Implement me"
      end
    end

    class Server < ServerBase
      def initialize(config_key_base)
        super()
        config.declare_dynamic_key("#{config_key_base}.port".intern,:uint,51337)
        config.declare_dynamic_key("#{config_key_base}.ssl".intern,:bool,true)
        config.declare_dynamic_key("#{config_key_base}.certificate".intern,:string,File.join(Ghun::Base::Blackboard.local.data_path,'visa','certs','server.crt'))
        config.declare_dynamic_key("#{config_key_base}.key".intern,:string,File.join(Ghun::Base::Blackboard.local.data_path,'visa','certs','server.key'))
        @port=config["#{config_key_base}.port".intern]
        @ssl=config["#{config_key_base}.ssl".intern]
        @ssl_config=Hash.new()
        if @ssl
          @ssl_config[:cert_file]=config["#{config_key_base}.certificate".intern]
          @ssl_config[:key_file]=config["#{config_key_base}.key".intern]
          @ssl_config[:cert]=File.read(@ssl_config[:cert_file]) unless @ssl_config[:cert_file].nil?()
          @ssl_config[:key]=File.read(@ssl_config[:key_file]) unless @ssl_config[:key_file].nil?()
          @ssl=!(@ssl_config[:cert].nil?() || @ssl_config[:key].nil?())
        end
      end

      def open_socket()
        tcp_server=::TCPServer.new(@port)
        if @ssl
          ssl_context=OpenSSL::SSL::SSLContext.new()
          ssl_context.cert=OpenSSL::X509::Certificate.new(@ssl_config[:cert])
          ssl_context.key=OpenSSL::PKey::RSA.new(@ssl_config[:key])
          ssl_server=OpenSSL::SSL::SSLServer.new(tcp_server,ssl_context)
          @server=ssl_server
        else
          @server=tcp_server
        end
      end

      def listen()
        open_socket unless defined?(@server) && @server
        super(@server)
      end
    end

    class VisaServer < Iof::Control::Server
      def initialize()
        super(:server)
        @controller=VisaController.new()
      end

      def handle_request(request,connection=nil)
        @controller.handle(request,connection)
      end

      def decode(line_in)
        Iof::Control::VisaRequest.decode(line_in)
      end

      def encode(response)
        Iof::Control::VisaResponse.encode(response)
      end
    end
  end
end
