# encoding: utf-8
# license: gpl3p

### BEGIN LICENSE NOTICE
# This file is part of the interconnect asset ontology framework (IO)
# Copyright (C) 2010 - 2013
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE NOTICE

### BEGIN AUTHOR LIST
#
### END AUTHOR LIST
require 'yaml'
require 'base64'
module Iof
  module Control
    Ghun::Log::Source.register_log_source("CONTROL")
    Ghun::Log::Type.register_log_type("SEND")
    Ghun::Log::Type.register_log_type("RECEIVE")

    class NoDeleteError < StandardError
    end

    class NoWriteError < StandardError
    end

    class UnknownTopologyIdError < StandardError
    end

    class TopologyExistsError < StandardError
    end

    class UnexpectedTransferEncodingError < StandardError
    end

    class UnexpectedMessageType < StandardError
    end

    class UnknownCommandError < StandardError
    end

    class CommandNotImplemented < StandardError
    end

    class MissingArgumentError < StandardError
    end

    class GenericProtocolError < StandardError
    end

    class InvalidRDFError < StandardError
    end

    class InvalidDataError < StandardError
    end

    class ConnectionError < StandardError
    end

    class CommandExecutionError < StandardError
    end

    class GenericError < StandardError
    end

    class UncleanGenericError < StandardError
    end

    class UnknownError < StandardError
    end


    class VisaRequest
      def initialize(command=nil,arguments=nil)
        @command=command
        @arguments=arguments
      end
      attr_accessor :command, :arguments

      def self.decode(line_in)
        tmp=VisaRequest.new()
        begin
          request=Base64.strict_decode64(line_in.chomp)
        rescue
          raise Iof::Control::UnexpectedTransferEncodingError, "Outer BASE64 encoding is corrupted or unknown encoding is used"
        end
        request.sub!(/!!de\.decoit\.visa\.protocol\./,'--- !ruby/object:Iof::Control::Visa')
        begin
          request=YAML::load(request)
        rescue
          raise Iof::Control::UnexpectedMessageType, "YAML encoding is corrupted or message type is unknown"
        end
        raise Iof::Control::UnexpectedMessageType, "#{request.class} in an unknown message type" unless request.is_a?(Iof::Control::VisaRequest)
        tmp.arguments=Hash.new()
        begin
          request.arguments.each do |key,value|
            tmp.arguments[Base64.strict_decode64(key).intern]=Base64.strict_decode64(value)
          end unless request.arguments.nil?()
          tmp.command=Base64.strict_decode64(request.command).intern
        rescue
          raise Iof::Control::UnexpectedTransferEncodingError, "Inner BASE64 encoding is corrupted or unknown encoding is used"
        end
        return tmp
      rescue
        raise
      end

      def self.encode(request)
        #used by development VISAClient, no need for dead certain exception handling
        tmp=VisaRequest.new()
        command_encoded=Base64.strict_encode64(request.command.to_s)
        arguments_encoded=Hash.new()
        request.arguments.each do |key,value|
          arguments_encoded[Base64.strict_encode64(key.to_s)]=Base64.strict_encode64(value)
        end unless request.arguments.nil?()
        tmp=VisaRequest.new(command_encoded,arguments_encoded)
        tmp_yaml=tmp.to_yaml()
        tmp_yaml.sub!(/--- !ruby\/object:Iof::Control::Visa/,'!!de.decoit.visa.protocol.')
        line_out=Base64.strict_encode64(tmp_yaml)
        return line_out
      end
    end

    class VisaResponse
      def initialize(return_code=0,data=nil,message="")
        @returnCode=return_code
        @data=data
        @message=message
      end
      attr_accessor :returnCode, :data, :message

      def return_code=(arg)
        @returnCode=arg
      end

      def return_code
        @returnCode
      end

      def self.decode(line_in)
        #used by development VISAClient, no need for dead certain exception handling
        tmp=VisaResponse.new()
        response=Base64.strict_decode64(line_in.chomp)
        response.sub!(/!!de\.decoit\.visa\.protocol\./,'--- !ruby/object:Iof::Control::Visa')
        response=YAML::load(response)
        tmp.return_code=response.return_code
        tmp.data=Hash.new()
        response.data.each do |key,value|
          tmp.data[Base64.strict_decode64(key).intern]=Base64.strict_decode64(value)
        end unless response.data.nil?()
        tmp.message=Base64.strict_decode64(response.message)
        return tmp
      end

      def self.encode(response)
        #used by VISAServer
        msg_encoded=Base64.strict_encode64(response.message)
        data_encoded=Hash.new()
        response.data.each do |key,value|
          data_encoded[Base64.strict_encode64(key.to_s)]=Base64.strict_encode64(value)
        end unless response.data.nil?()
        tmp=VisaResponse.new(response.return_code,data_encoded,msg_encoded)
        tmp_yaml=tmp.to_yaml()
        tmp_yaml.sub!(/--- !ruby\/object:Iof::Control::Visa/,'!!de.decoit.visa.protocol.')
        line_out=Base64.strict_encode64(tmp_yaml)
        return line_out
      rescue => e
        #returning nil in case of an error would trigger the regular error case handling,
        #but may cause a endless loop, when failing during encoding an error message
        Ghun::Base::Blackboard.logger.error "Failed to encode response #{response.inspect()}", e
        return ""
      end

      def self.translate_from_exception(exception)
        result=nil
        case exception
        when UnknownTopologyIdError
          result=VisaResponse.new(100,{},"#{exception.message}")
        when TopologyExistsError
          result=VisaResponse.new(101,{},"#{exception.message}")
        when InvalidRDFError
          result=VisaResponse.new(102,{},"#{exception.message}")
        when InvalidDataError
          result=VisaResponse.new(103,{},"#{exception.message}")
        when ConnectionError
          result=VisaResponse.new(106,{},"#{exception.message}")
        when CommandExecutionError
          result=VisaResponse.new(106,{},"#{exception.message}")
        when NoWriteError
          result=VisaResponse.new(106,{},"#{exception.message}")
        when NoDeleteError
          result=VisaResponse.new(107,{},"#{exception.message}")
        when UnexpectedTransferEncodingError
          result=VisaResponse.new(180,{},"#{exception.message}")
        when UnexpectedMessageType
          result=VisaResponse.new(181,{},"#{exception.message}")
        when UnknownCommandError
          result=VisaResponse.new(182,{},"#{exception.message}")
        when CommandNotImplemented
          result=VisaResponse.new(183,{},"#{exception.message}")
        when MissingArgumentError
          result=VisaResponse.new(184,{},"#{exception.message}")
        when GenericProtocolError
          result=VisaResponse.new(189,{},"#{exception.message}")
        when UncleanGenericError
          result=VisaResponse.new(198,{},"#{exception.message}")
        when GenericError
          result=VisaResponse.new(199,{},"#{exception.message}")
        when UnknownError
          result=VisaResponse.new(255,{},"#{exception.message}")
        else
          result=self.error_message_response
        end
        return result
      rescue
        self.error_message_response
      end


      def self.error_message_response(message="An error occured")
        VisaResponse.new(255,{},message)
      end

      def self.successful_data_response(data,return_code=0,message="")
        VisaResponse.new(return_code,data,message)
      end
    end
  end
end
