# encoding: utf-8
# license: gpl3p

### BEGIN LICENSE NOTICE
# This file is part of the interconnect asset ontology framework (IO)
# Copyright (C) 2010 - 2013
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE NOTICE

### BEGIN AUTHOR LIST
#
### END AUTHOR LIST
require 'net/scp'
module Iof
  module Control

    CommandArgumentDescription=Struct.new(:name,:optional,:default)

    class ControllerBase < Ghun::Base::Module
      def initialize(commands,command_arguments,command_aliases)
        super(Ghun::Log::Source::CONTROL)
        @commands=commands || []
        @command_arguments=command_arguments || {}
        @command_aliases=command_aliases || {}
      end

      def handle(request,connection)
        raise Ghun::Base::AbstractClassError, "Implement me"
      end

    private

      def execute(command,arguments,connection)
        result=nil
        c=nil
        args=nil
        #preflight check
        begin
          c=resolv_command_alias(command)
          debug "Executing command #{c}"
          debug "Command #{c} was called with alias name #{command}" unless c==command
          method="execute_#{c.to_s}".intern
          raise Iof::Control::UnknownCommandError, "Command #{command} is not defined" unless is_command?(c)
          raise Iof::Control::CommandNotImplemented, "Command #{command} is not implemented yet" unless self.respond_to?(method, true)
          args=check_command_arguments(c, arguments)
          args[:connection]=connection
          raise Iof::Control::GenericError, "Failed to prepare command execution" if c.nil?() || args.nil?()
          result=self.send(method, args)
          return result
        rescue => e
          error "Failed to execute command #{command}", e
          result=Iof::Control::VisaResponse.translate_from_exception(e)
        end
        return result
      end

      def sanitize_command_name(command)
        new_command=command.to_s
        new_command.tr!('^A-Za-z0-9_-','')
        raise Iof::Control::UnknownCommandError, "Unknown command '#{command}'" if new_command==""
        return new_command.intern
      rescue Iof::Control::UnknownCommandError => e
        raise e
      rescue => e
        error "Failed to sanitize command '#{command}'", e
        raise Iof::Control::UnknownCommandError, "Unknown command '#{command}'"
      end

      def resolv_command_alias(command)
        c=sanitize_command_name(command)
        if @command_aliases.include?(c)
          return @command_aliases[c]
        else
          return c
        end
      rescue Iof::Control::UnknownCommandError => e
        raise e
      rescue => e
        error "Failed to unalias command '#{command}'", e
        raise Iof::Control::UnknownCommandError, "Unknown command '#{command}'"
      end

      def is_command?(command)
        @commands.include?(resolv_command_alias(command))
      end

      def check_command_arguments(command,arguments)
        c=resolv_command_alias(command)
        a=@command_arguments[c]
        result={}
        if a.nil?() || a.empty?()
          #no arguments defined
          warn "No arguments definition found for command #{c}; assuming empty argument set" if a.nil?()
          debug "Found #{arguments.keys.size} undefined arguments for command '#{command}': '#{arguments.keys.to_s}'" unless arguments.nil?() || arguments.keys.empty?()
          return result
        elsif arguments.nil?()
          error "Got no argument hash for command '#{command}'"
          raise Iof::Control::GenericProtocolError, "Argument hash MUST NOT be empty; please check your protocol implementation"
        else
          done=[]
          a.each do |argument|
            if arguments.include?(argument[:name])
              #use provided name; optional and default do not matter
              done << argument[:name]
              result[argument[:name]]=arguments[argument[:name]].dup() unless arguments[argument[:name]].nil?()
            elsif !argument[:optional] && !argument[:default].nil?()
              #argument is missing and mandatory => use defined default
              done << argument[:name]
              result[argument[:name]]=argument[:default]
            elsif argument[:optional]
              #argument is missing and optional => use defined default
              #default value for argument[:default] is nil
              done << argument[:name]
              result[argument[:name]]=argument[:default]
            else
              #remaining case: argument is missing, mandatory and has no default => Fail
              raise Iof::Control::MissingArgumentError, "Argument '#{argument[:name]}' is missing for command '#{command}'"
            end
          end
          unused=arguments.keys()-done
          debug "Found #{unused.size} undefined arguments: '#{unused.to_s}'" unless unused.empty?()
        end
        return result
      rescue => e
        error "Failed to check arguments for command '#{command}'", e
        raise e
      end
    end

    class VisaController < ControllerBase
      Ghun::Base::Blackboard.config.declare_key(:"visa.safe_mode",:bool,true)
      Ghun::Base::Blackboard.config.declare_key(:"visa.paranoia_mode",:bool,true)
      Ghun::Base::Blackboard.config.declare_key(:"visa.demonstrator_mode",:bool,false)

      Ghun::Base::Blackboard.config.declare_key(:"visa.topology_dir",:string,File.join(Ghun::Base::Blackboard.local.data_path,'visa','topologies'))
      Ghun::Base::Blackboard.config.declare_key(:"visa.image_dir",:string,File.join(Ghun::Base::Blackboard.local.data_path,'visa','images'))
      Ghun::Base::Blackboard.config.declare_key(:"visa.script_dir",:string,File.join(Ghun::Base::Blackboard.local.data_path,'visa','scripts'))

      Ghun::Base::Blackboard.config.declare_key(:"visa.source.server",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.system_user",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.system_password",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.os_tenant",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.os_user",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.os_password",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.os_url",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.os_service_endpoint",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.os_service_token",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.topology_name",:string,"")

      Ghun::Base::Blackboard.config.declare_key(:"visa.source.reverse_connection_host",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.reverse_connection_user",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.source.reverse_connection_key",:string,File.join(Ghun::Base::Blackboard.local.data_path,'visa','certs','visa'))

      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.server",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.system_user",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.system_password",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.os_tenant",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.os_user",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.os_password",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.os_url",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.os_service_endpoint",:string)
      Ghun::Base::Blackboard.config.declare_key(:"visa.destination.os_service_token",:string)

      include Ghun::Base::Configuration
      @@topology_mutex=Monitor.new()
      @@visa_mutex=Monitor.new()
      @@commands=[
        :get_topology_list,
        :get_topology,
        :store_topology,
        :drop_topology,
        :collect,
        :replicate,
        :cleanup,
        :reset,
        :disconnect,
        ]

      @@command_aliases={
        :get_topo_list  => :get_topology_list,
        :get_topo       => :get_topology,
        :write_topo     => :store_topology,
        :bye            => :disconnect,
      }

      @@command_arguments={
        :get_topology_list  => [],
        :get_topology       => [CommandArgumentDescription.new(:topology)],
        :store_topology     => [CommandArgumentDescription.new(:topology),
                                CommandArgumentDescription.new(:data),
                                CommandArgumentDescription.new(:description,true)],
        :drop_topology      => [CommandArgumentDescription.new(:topology)],
        :collect            => [CommandArgumentDescription.new(:topology),
                                CommandArgumentDescription.new(:description,true)],
        :replicate          => [CommandArgumentDescription.new(:topology)],
        :cleanup            => [],
        :reset              => [],
        :disconnect         => [],
      }

      def initialize()
        super(@@commands,@@command_arguments,@@command_aliases)
        init_configuration()
        @topology_dir=config[:"visa.topology_dir"]
        @script_dir=config[:"visa.script_dir"]
        @image_dir=config[:"visa.image_dir"]
        @safe=config[:"visa.safe_mode"]
        @paranoid=config[:"visa.paranoia_mode"]
        @demonstrator=config[:"visa.demonstrator_mode"]
        @safe=@safe || @paranoid
        Iof::Data::Mapper.instance
        init_topology_list()
        init_translation()
      end

      def handle(request,connection)
        @@visa_mutex.synchronize {
          begin
            response=nil
            raise Iof::Control::UnexpectedMessageType, "#{request.class} in an unknown message type" unless request.is_a?(Iof::Control::VisaRequest)
            response=execute(request.command, request.arguments,connection)
            return response
          rescue => e
            error "An error occured while handling request #{request.inspect}", e
            return Iof::Control::VisaResponse.translate_from_exception(e)
          end
        }
      end

    private

      def execute_get_topology_list(arguments)
        result=Hash.new()
        @topologies.each do |key,value|
          result[key]=value[:description]
        end
        return VisaResponse.successful_data_response(result)
      rescue => e
        error "An error occured while handling get_topo_list request", e
        return VisaResponse.translate_from_exception(e)
      end

      def execute_get_topology(arguments)
        topology=arguments[:topology]
        result=Hash.new()
        @@topology_mutex.synchronize {
          raise UnknownTopologyIdError, "Topology #{topology} is unknown" unless @topologies.include?(topology.intern)
          result[topology.intern]=@topologies[topology.intern][:topology] if @topologies.include?(topology.intern)
          result[:os_translation]=@topologies[topology.intern][:mapping] if @topologies.include?(topology.intern) && @topologies[topology.intern].include?(:mapping)
        }
        return VisaResponse.successful_data_response(result)
      rescue => e
        error "An error occured while handling get_topo request for topology '#{topology}'", e
        return VisaResponse.translate_from_exception(e)
      end

      def execute_store_topology(arguments)
        topology=arguments[:topology]
        data=arguments[:data]
        description=arguments[:topology]
        description=arguments[:description] unless arguments[:description].nil?()

        @@topology_mutex.synchronize {
          top=sanitize_name(topology)
          if @topologies.include?(topology.intern)
            debug "Updating topology '#{topology}'"
            t=@topologies[topology.intern]
            raise Iof::Control::NoWriteError, "Topology #{topology} is protected from being overwritten" if t[:no_write] || @paranoid
            store_topology(true, topology, data, description, t[:ontology_file], t[:topology_file])
          elsif @topologies.include?(top.intern)
            debug "Updating topology '#{top}'"
            t=@topologies[top.intern]
            raise Iof::Control::NoWriteError, "Topology #{top} is protected from being overwritten" if t[:no_write] || @paranoid
            store_topology(true, top, data, description, t[:ontology_file], t[:topology_file])
          else
            raise Iof::Control::NoWriteError, "Topology #{top} is protected from being overwritten" if @paranoid
            debug "Creating topology '#{topology}' with name '#{top}' and description '#{description}'"
            d=File.join(@topology_dir,top)
            if File.exists?(d) then
              raise Iof::Control::UncleanGenericError, "Data structure for topology #{top} exists, but may be corrupted"
            else
              FileUtils.mkdir(d)
              store_topology(false, top, data, description, File.join(d,"#{top}.owl"), File.join(d,"#{top}.rdf"))
            end
          end
        }
        return VisaResponse.successful_data_response(nil)
      rescue => e
        error "An error occured while handling write_topo request for topology '#{topology}'", "Data #{data}", e
        return VisaResponse.translate_from_exception(e)
      end

      def execute_drop_topology(arguments)
        topology=sanitize_name(arguments[:topology])
        if @topologies.include?(topology.intern) && @topologies[topology.intern].include?(:topology_dir) && !@topologies[topology.intern][:topology_dir]
          begin
            t=@topologies[topology.intern]
            raise Iof::Control::NoDeleteError, "Topology #{topology} is protected from being deleted" if t[:no_delete] || @safe
            FileUtils.rm_r(t[:topology_dir])
            @topologies.delete(topology.intern)
          rescue Iof::Control::NoDeleteError => e
            error e
            raise e
          rescue => e
            error "Failed to drop topology #{topology}", e
            raise UncleanGenericError, "Failed to drop topology #{topology}"
          end
        elsif File.exists?(File.join(@topology_dir,topology))
          begin
            raise Iof::Control::NoDeleteError, "Topology #{topology} is protected from being deleted" if File.directory?(File.join(@topology_dir,topology)) && File.exists?(File.join(@topology_dir,topology,'.no_delete')) || @safe
            FileUtils.rm_r(File.join(@topology_dir,topology))
          rescue Iof::Control::NoDeleteError => e
            error e
            raise e
          rescue => e
            error "Failed to drop topology #{topology}", e
            raise UncleanGenericError, "Failed to drop topology #{topology}"
          end
        else
          raise Iof::Control::UnknownTopologyIdError, "Topology #{topology} is unknown"
        end
        @topologies.delete(topology.intern) if @topologies.include?(topology.intern)
        return VisaResponse.successful_data_response(nil)
      end

      def execute_reset(arguments)
        arguments[:connection].puts(Iof::Control::VisaResponse.encode(VisaResponse.new(254))) unless arguments[:connection].nil?()
        error_count=0
        intended_error_count=0
        dirs=Dir.glob(File.join(@topology_dir,'*'))
        dirs.each do |d|
          begin
            execute_drop_topology({:topology => d.sub(/.*\//,'')})
          rescue Iof::Control::NoDeleteError => e
            intended_error_count+=1
          rescue => e
            error "Failed to delete topolgy #{d.sub(/.*\//,'')}", e
            error_count+=1
          end
        end
        begin
          execute_cleanup({})
        rescue => e
          error "Failed to cleanup replication environment", e
          error_count+=1
        end
        if error_count>0 && intended_error_count>0
          raise Iof::Control::GenericError, "#{error_count} unexpected errors and #{intended_error_count} protected topologies occured during cleanup execution"
        elsif error_count>0
          raise Iof::Control::GenericError, "#{error_count} unexpected errors"
        elsif intended_error_count>0
          return VisaResponse.successful_data_response(nil,1,"#{intended_error_count} protected topologies occured during cleanup execution")
        end
        return VisaResponse.successful_data_response(nil)
      end

      def execute_cleanup(arguments)
        arguments[:connection].puts(Iof::Control::VisaResponse.encode(VisaResponse.new(254))) unless arguments[:connection].nil?()
        if @safe
          sleep(30)
          return VisaResponse.successful_data_response(nil,0,"Cleanup is disabled in safe or paranoia mode")
        end

        script=File.read(File.join(@script_dir,"cleanup_openstack.sh"))
        credentials="export OS_TENANT_NAME=#{config[:"visa.destination.os_tenant"]}\nexport OS_USERNAME=#{config[:"visa.destination.os_user"]}\nexport OS_PASSWORD=#{config[:"visa.destination.os_password"]}\nexport OS_AUTH_URL=#{config[:"visa.destination.os_url"]}\n"
        script.sub!(/%CREDENTIALS%/,credentials)
        script.sub!(/%DEMONSTRATOR%/,@demonstrator.to_s)
        command="#{script}"
        auth=Ghun::Auth::Item.new()
        auth.username=config[:"visa.destination.system_user"]
        auth.password=config[:"visa.destination.system_password"]
        begin
          connection=Ghun::Remote::PlainSSH.new(config[:"visa.destination.server"],auth)
        rescue => e
          error "Could not connect to destination environment", e
          raise ConnectionError, "Could not connect to destination environment"
        end
        begin
          result=connection.run_command(command)
        rescue => e
          error "Could not cleanup destination environment", e
          raise CommandExecutionError, "Could not cleanup destination environment"
        end
        begin
          connection.disconnect()
        rescue => e
          error "Could not disconnect from destination environment", e
          raise ConnectionError, "Could not disconnect from destination environment"
        end
        debug result
        return VisaResponse.successful_data_response(nil)
      rescue => e
        error "Failed to clean up destination environment",e
        return VisaResponse.translate_from_exception(e)
      end

      def prepare_replication_data(rdf_data)
        topology=Iof::Query::VisaRDF.import(rdf_data)
        result=Hash.new()
        result[:number_of_vms]=0
        result[:number_of_networks]=0
        result[:vms]=Array.new()
        result[:images]=Array.new()
        result[:source_images]=Array.new()
        result[:networks]=Array.new()
        result[:vm_to_network]=Array.new()
        result[:expected]=Hash.new()
        topology.each do |v|
          v[:devices].each do |d|
            next unless d[:type]==:host
            unless result[:vms].include?(d[:internal_name])
              result[:vms] << d[:internal_name]
              result[:number_of_vms]+=1
              if d[:image_file].nil?() || d[:image_file]==''
                result[:images] << "dummy"
                result[:source_images] << File.join(config[:"visa.image_dir"],"dummy")
              else
                result[:images] << d[:image_file].sub(/.*\//,'')
                result[:source_images] << File.join(config[:"visa.image_dir"],d[:image_file])
              end
              networks=Array.new()
              d[:interfaces].each do |i|
                i[:addresses].each do |a|
                  next unless a[:type]==:ipv4
                  if a[:network].nil?()
                    addr=::IPAddress::IPv4.new(a[:value])
                    t_a=nil
                    n=nil
                    t_n=nil
                    @address_translation.each do |translation|
                      next unless translation[:ipv4]
                      next unless translation[:original].include?(addr)
                      n=Iof::Query::Network.new(:ipv4,translation[:original].to_s,translation[:original].prefix.to_s,nil)
                      t_n=translate_network(n)
                      t_a=translate_address(a)
                      result[:expected][d[:internal_name].intern]=Array.new() unless result[:expected].include?(d[:internal_name].intern)
                      result[:expected][d[:internal_name].intern] << {:address => a, :network => n, :translated_address => t_a, :translated_network => t_n}
                    end
                  else
                    n=translate_network(a[:network])
                    unless result[:networks].include?("#{n[:value]}/#{n[:netmask_length]}")
                      result[:number_of_networks]+=1
                      result[:networks] << "#{n[:value]}/#{n[:netmask_length]}"
                    end
                    result[:expected][d[:internal_name].intern]=Array.new() unless result[:expected].include?(d[:internal_name].intern)
                    result[:expected][d[:internal_name].intern] << {:address => a, :network => a[:network], :translated_address => translate_address(a), :translated_network => n}
                    networks << result[:networks].index("#{n[:value]}/#{n[:netmask_length]}")
                  end
                end unless i[:addresses].nil?()
              end unless d[:interfaces].nil?()
              result[:vm_to_network] << networks.join(' ')
            end
          end unless v[:devices].nil?()
        end unless topology.nil?()
        return result
      end

      def postprocess_replication_result(topology,replication_result,expected)
        assigned=Hash.new()
        mapping=Hash.new()
        garbage=true
        replication_result[:stdout].each_line do |line|
          if line.match(/^==*$/)
            garbage=false
            next
          elsif garbage
            next
          else
            host,address=line.split(';')
            next if host.nil?() || host=='' || address.nil?() || address==''
            host.strip!
            address.strip!
            address.sub!(/public=/,'')
            next if address==''
            addresses=address.split(',')
            addresses.each { |a| a.strip! }
            assigned[host.intern]=addresses
          end
        end
        assigned.each do |host,addresses|
          #host existed but was not replicated
          next unless expected.include?(host)
          ex=expected[host]
          addresses.each do |a|
            ex.each do |e|
              expected_net=::IPAddress::IPv4.new("#{e[:translated_network][:value]}/#{e[:translated_network][:netmask_length]}")
              mapping[e[:address][:value].intern]=a if expected_net.include?(::IPAddress::IPv4.new(a))
            end
          end
        end

        #map to '' for not replicated hosts and missing addresses
        expected.each do |host,address_network|
          address_network.each do |a|
            mapping[a[:address][:value].intern]='' unless mapping.include?(a[:address][:value].intern)
          end
        end

        File.open(@topologies[topology.intern][:mapping_file],"w") do |f|
          mapping.each do |old,new|
            f.puts("#{old.to_s};#{new.to_s};")
          end
        end
        @topologies[topology.intern][:mapping]=File.read(@topologies[topology.intern][:mapping_file]).chomp
      end


      def execute_replicate(arguments)
        arguments[:connection].puts(Iof::Control::VisaResponse.encode(VisaResponse.new(254))) unless arguments[:connection].nil?()
        if @safe
          sleep(30)
          return VisaResponse.successful_data_response(nil,0,"Replicate is disabled in safe or paranoia mode")
        end

        topology=sanitize_name(arguments[:topology])
        raise UnknownTopologyIdError, "Topology #{arguments[:topology]} (#{topology}) is unknown" unless @topologies.include?(topology.intern)

        script=File.read(File.join(@script_dir,"replicate_openstack.sh"))
        credentials="export OS_TENANT_NAME=#{config[:"visa.destination.os_tenant"]}\nexport OS_USERNAME=#{config[:"visa.destination.os_user"]}\nexport OS_PASSWORD=#{config[:"visa.destination.os_password"]}\nexport OS_AUTH_URL=#{config[:"visa.destination.os_url"]}\n"
        script.sub!(/%CREDENTIALS%/,credentials)
        replication_data=prepare_replication_data(@topologies[topology.intern][:topology])

        number_of_vms=replication_data[:number_of_vms]
        number_of_networks=replication_data[:number_of_networks]
        vms=""
        images=""
        networks=""
        vm_to_network=""
        replication_data[:vms].each_with_index {|v,i| vms+="vms[#{i}]='#{v}'\n"}
        replication_data[:images].each_with_index {|v,i| images+="images[#{i}]='#{v}'\n"}
        replication_data[:networks].each_with_index {|v,i| networks+="networks[#{i}]='#{v}'\n"}
        replication_data[:vm_to_network].each_with_index {|v,i| vm_to_network+="vm_to_network[#{i}]='#{v}'\n"}

        script.sub!(/%NUMBER_OF_VMS%/,number_of_vms.to_s)
        script.sub!(/%NUMBER_OF_NETWORKS%/,number_of_networks.to_s)
        script.sub!(/%FILL_VMS%/,vms)
        script.sub!(/%FILL_IMAGES%/,images)
        script.sub!(/%FILL_NETWORKS%/,networks)
        script.sub!(/%FILL_VM_TO_NETWORK%/,vm_to_network)

        script.sub!(/%DEMONSTRATOR%/,@demonstrator.to_s)


        auth=Ghun::Auth::Item.new()
        auth.username=config[:"visa.destination.system_user"]
        auth.password=config[:"visa.destination.system_password"]
        if @demonstrator
          info "Skipping image transfer"
        else
          begin
            connection=Ghun::Remote::PlainSSH.new(config[:"visa.destination.server"],auth)
          rescue => e
            error "Could not connect to destination environment", e
            raise ConnectionError, "Could not connect to destination environment"
          end
          begin
            connection.run_command("mkdir -p images")
          rescue => e
            error "Could not cleanup destination environment", e
            raise CommandExecutionError, "Could not cleanup destination environment"
          end
          begin
            connection.disconnect()
          rescue => e
            error "Could not disconnect from destination environment", e
            raise ConnectionError, "Could not disconnect from destination environment"
          end
          begin
            replication_data[:source_images].each do |i|
              Net::SCP.upload!(config[:"visa.destination.server"], config[:"visa.destination.system_user"],i,"images/", :ssh => { :password => config[:"visa.destination.system_password"] })
            end
          rescue => e
            error "Could not connect to destination environment", e
            raise ConnectionError, "Could not connect to destination environment"
          end
        end
        command="#{script}"
        begin
          connection=Ghun::Remote::PlainSSH.new(config[:"visa.destination.server"],auth)
        rescue => e
          error "Could not connect to destination environment", e
          raise ConnectionError, "Could not connect to destination environment"
        end
        begin
          result=connection.run_command(command)
        rescue => e
          error "Could not cleanup destination environment", e
          raise CommandExecutionError, "Could not cleanup destination environment"
        end
        begin
          connection.disconnect()
        rescue => e
          error "Could not disconnect from destination environment", e
          raise ConnectionError, "Could not disconnect from destination environment"
        end
        debug result
        postprocess_replication_result(topology, result, replication_data[:expected])
        return VisaResponse.successful_data_response(nil)
      rescue => e
        error "Failed to replicate topology #{topology}", e
        return VisaResponse.translate_from_exception(e)
      end

      def execute_collect(arguments)
        arguments[:connection].puts(Iof::Control::VisaResponse.encode(VisaResponse.new(254))) unless arguments[:connection].nil?()
        if @safe
          sleep(30)
          return VisaResponse.successful_data_response(nil,0,"Collect is disabled in safe or paranoia mode")
        end
        if @demonstrator
          sleep(5)
          return VisaResponse.successful_data_response(nil,0,"Collect is disabled in demonstrator mode")
        end

        topology=sanitize_name(arguments[:topology])
        raise Iof::Control::TopologyExistsError ,"Topology #{topology} exists" if @topologies.include?(topology.intern)
        raise Iof::Control::TopologyExistsError ,"Topology #{topology} exists" if File.exists?(File.join(@topology_dir,topology))
        description=arguments[:topology]
        description=arguments[:description] unless arguments[:description].nil?()
        config.set_key(:"visa.source.topology_name", topology.to_s)
        dir=File.join(@topology_dir,topology.to_s)
        conf_dir=File.join(dir,"config")
        FileUtils.mkdir(dir)

        FileUtils.mkdir(conf_dir)
        begin
          collector=Iof::Runner::Base.new()
          collector.run()
        rescue => e
          error "Failed to collect from source environment", e
          raise ConnectionError, "Failed to collect from source environment; see logfiles for details"
        end
        FileUtils.cp(File.join(Ghun::Base::Blackboard.local.data_path,'ontologies','network.owl'),File.join(dir,"#{topology.to_s}.owl"))
        FileUtils.cp(File.join(Ghun::Base::Blackboard.application.data_path,'ontologies','network.owl'),File.join(Ghun::Base::Blackboard.local.data_path,'ontologies','network.owl'))
        File.open(File.join(dir,"description"),"w") { |f| f.puts(description)}
        init_topology_list
        config.set_key(:"visa.source.topology_name", "")
        VisaResponse.successful_data_response(nil)
      rescue => e
        error "Failed to collect configuration from source environment", e
        VisaResponse.translate_from_exception(e)
      end

      def execute_disconnect(arguments)
        return false
      end

      def sanitize_name(name)
        new=""
        new=name.tr(' ','_') unless name.nil?()
        new=name.tr('.','_') unless name.nil?()
        new=name.tr('\'','_') unless name.nil?()
        new=name.tr('"','_') unless name.nil?()
        new.tr!('^ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_-','')
        new=SecureRandom.uuid() if new==""
        return new
      end

      def init_topology_list()
        @@topology_mutex.synchronize {
          @topologies=Hash.new()
          dirs=Dir.glob(File.join(@topology_dir,'*'))
          dirs.each do |d|
            next unless File.directory?(d)
            begin
              t=Hash.new()
              t[:valid]=true
              t[:name]=d.sub(/.*\//,'')
              if File.exists?(File.join(d,'description'))
                t[:description]=File.read(File.join(d,'description')).chomp()
              else
                t[:description]=t[:name].dup()
              end
              t[:topology_file]=File.join(d,"#{t[:name]}.rdf")
              t[:mapping_file]=File.join(d,"#{t[:name]}.osm")
              t[:topology_dir]=d
              t[:ontology_file]=File.join(d,"#{t[:name]}.owl")
              t[:csv_file]=File.join(d,"#{t[:name]}.csv")

              t[:mapping]=File.read(t[:mapping_file]).chomp if File.exists?(t[:mapping_file])
              if !(File.exists?(t[:ontology_file]) || File.exists?(t[:ontology_file])) && File.exists?(t[:csv_file])
                debug "Creating RDF file from CSV file"
                convert_csv_to_rdf(t[:name],t[:csv_file],t[:topology_file])
              end
              if File.exists?(t[:ontology_file])
                t[:ontology_file]=t[:ontology_file]
                debug "Found OWL file #{t[:ontology_file]}"
              elsif File.exists?(t[:topology_file])
                warn "Missing OWL file for topology #{t[:name]}, but RDF file exists"
                t[:no_owl]=true
              else
                error "Missing OWL file for topology #{t[:name]}"
                t[:valid]=false
              end
              t[:topology_file]=File.join(d,"#{t[:name]}.rdf")
              if File.exists?(t[:topology_file])
                if t.include?(:no_owl)
                  debug "Creating OWL from RDF file #{t[:topology_file]}"
                  t[:topology]=File.read(t[:topology_file]).chomp()
                  store_topology(false, t[:name], t[:topology], t[:description], t[:ontology_file], t[:topology_file])
                  t[:valid]=false if !File.exists?(t[:ontology_file]) || !File.exists?(t[:topology_file]) || t[:topology].nil?() || t[:topology]==""
                else
                  debug "Found RDF file #{t[:topology_file]}"
                  t[:topology]=File.read(t[:topology_file]).chomp()
                end
              else
                unless t.include?(:no_owl) || t[:no_owl]
                  debug "Generating topology #{t[:name]} from #{t[:ontology_file]}"
                  t[:topology]=generate_topology(t[:name],t[:topology_file],t[:ontology_file])
                end
              end if t[:valid]
              t[:no_delete]=File.exists?(File.join(d,".no_delete"))
              t[:no_write]=File.exists?(File.join(d,".no_write"))
              t[:valid]=false if !File.exists?(t[:ontology_file]) || !File.exists?(t[:topology_file]) || t[:topology].nil?() || t[:topology]==""
              check_for_loop(t[:name], t[:topology]) if t[:valid]
            rescue => e
              error "Failed to load topology #{d}", e
            end
            if t[:valid] then
              info "Successfully loaded topology #{t[:name]} (#{t[:description].inspect})"
              @topologies[t[:name].intern]=t
            else
              warn "Directory structure #{d} is invalid, ignoring topology"
            end
          end
        }
      end

      def generate_topology(name,rdf_file,ontology_file)
        debug "Generating topology #{name} from #{ontology_file}"
        query=Iof::Query::Openstack.new(ontology_file)
        topology=Iof::Query::VisaRDF.export(query.query_topology(),rdf_file)
        query.finish(true)
        return topology[:output]
      rescue => e
        error "Failed to generated topology from #{ontology_file}", e
        return nil
      end

      def store_topology(update, topology, data, description, ontology_file, topology_file)
        if update
          update_topology(topology, data, ontology_file, topology_file)
        else
          write_new_topology(topology, data, ontology_file, topology_file)
        end
        if description.nil?() && @topologies[topology.intern][:description]
          #description for topology keeps unchanged
        elsif description.nil?() && @topologies[topology.intern][:description].nil?()
          #no description given and none known, using topology as description
          @topologies[topology.intern][:description]=topology
        else
          @topologies[topology.intern][:description]=description
        end
        @topologies[topology.intern][:topology_dir]=File.join(@topology_dir,topology) unless @topologies[topology.intern].include?(:topology_dir)
        File.open(File.join(@topology_dir,topology,"description"),"w") { |f| f.puts(@topologies[topology.intern][:description]) }
      end

      def update_topology(topology, data, ontology_file, topology_file)
        now=Time.now()
        File.open("#{topology_file}.incoming.#{now.to_f}.rdf","w") { |f| f.puts(data) }
        #Move ontology file to force write instead of update
        FileUtils.mv(ontology_file,"#{ontology_file}.before_change.#{now.to_f}.owl")
        FileUtils.cp(topology_file,"#{topology_file}.before_change.#{now.to_f}.rdf")

        begin
          import=Iof::Query::VisaRDF.import(data)
        rescue => e
          error "Failed to parse RDF structure", e
          raise InvalidRDFError, "RDF parsing for topology #{topology} failed"
        end

        data_handler=@data_handler=Iof::Data::Handler.new(true,ontology_file)

        begin
          parser=Iof::Parser::VisaRDF.new(data_handler)
          parser.update(import)
          data_handler.write_changes(false)
        rescue => e
          error "Failed to insert RDF data into ontology", e
          raise InvalidDataError, "RDF to ontology for topology #{topology} failed"
        end

        FileUtils.cp(ontology_file,"#{ontology_file}.after_change.#{now.to_f}.owl")
        new_data=generate_topology(topology, topology_file, ontology_file)
        @topologies[topology.intern][:topology]=new_data
        File.open(topology_file,"w") { |f| f.puts(new_data) }
        File.open("#{topology_file}.after_change.#{now.to_f}.rdf","w") { |f| f.puts(new_data) }
        result=check_for_loop(topology, new_data)
        raise InvalidDataError, "Topology #{topology} contains loop on layer 2; see logfile for details" if result==:layer2
      rescue => e
        error "Failed to update topology #{topology}", e
        raise e
      end

      def write_new_topology(topology, data, ontology_file, topology_file)
        now=Time.now()
        t=Hash.new()
        t[:name]=topology
        t[:description]=topology
        t[:topology_file]=topology_file
        t[:ontology_file]=ontology_file
        File.open("#{topology_file}.initial.#{now.to_f}.rdf","w") { |f| f.puts(data) }
        begin
          import=Iof::Query::VisaRDF.import(data)
        rescue => e
          error "Failed to parse RDF structure", e
          raise InvalidRDFError, "RDF parsing for topology #{topology} failed"
        end

        data_handler=@data_handler=Iof::Data::Handler.new(true,t[:ontology_file])

        begin
          parser=Iof::Parser::VisaRDF.new(data_handler)
          parser.write(import)
          data_handler.write_changes(false)
        rescue => e
          error "Failed to insert RDF data into ontology", e
          raise InvalidDataError, "RDF to ontology for topology #{topology} failed"
        end

        FileUtils.cp(ontology_file,"#{ontology_file}.after_change.#{now.to_f}.owl")
        new_data=generate_topology(topology, topology_file, ontology_file)
        t[:topology]=new_data
        File.open(topology_file,"w") { |f| f.puts(new_data) }
        File.open("#{topology_file}.after_change.#{now.to_f}.rdf","w") { |f| f.puts(new_data) }
        t[:valid]=true
        @topologies[topology.intern]=t
        result=check_for_loop(topology, new_data)
        raise InvalidDataError, "Topology #{topology} contains loop on layer 2; see logfile for details" if result==:layer2
      rescue => e
        error "Failed to write topology #{topology}", e
        raise e
      end

      def convert_csv_to_rdf(vsa,csv,rdf)
        raw=File.read(csv)
        result=Iof::Query::VisaRDF.import_from_csv(vsa, raw)
        Iof::Query::VisaRDF.export(result, rdf)
      end

      def init_translation
        @address_translation=Array.new()
#        @address_translation << {:original_network => "172.31.252.0/24", :translated_network => "192.168.21.0/24", :action => :translate}
#        @address_translation << {:original_network => "172.31.253.0/24", :translated_network => "192.168.20.0/24", :action => :translate}
#        @address_translation << {:original_network => "10.240.0.0/24", :translated_network => "10.42.0.0/24", :action => :translate}
        @address_translation.each do |translation|
          unless translation[:original_network].nil?() || translation[:original_network]=="" then
            translation[:original]=::IPAddress.parse(translation[:original_network])
          else
            translation[:action]=:none
          end
          unless translation[:translated_network].nil?() || translation[:translated_network]=="" then
            translation[:translated]=::IPAddress.parse(translation[:translated_network])
          else
            translation[:action]=:none
          end
          unless translation[:action]==:none || translation[:translated].ipv6?()==translation[:original].ipv6?()
            warn "Address translation from IPv4 to IPv6 or vice versa is not supported"
            translation[:action]=:none
          end
          unless translation[:action]==:none || translation[:translated].prefix==translation[:original].prefix
            warn "Address translation is only supported for equally sized subnets"
            translation[:action]=:none
          end
          unless translation[:action]==:none then
            if translation[:translated].ipv4?() then
              translation[:diff]=translation[:translated].to_u32-translation[:original].to_u32
              translation[:ipv4]=true
              translation[:ipv6]=false
            else
              translation[:diff]=translation[:translated].to_u128-translation[:original].to_u128
              translation[:ipv4]=false
              translation[:ipv6]=true
            end
            translation[:translated_prefix]=translation[:translated].prefix
          end
        end
      end

      def translate_address_value(address)
        a=::IPAddress.parse(address)
        new=nil
        @address_translation.each do |translation|
          next if translation[:action]==:none
          if a.ipv4?() then
            next unless translation[:ipv4]
            next unless translation[:original].include?(a)
            new=::IPAddress::IPv4.parse_u32((a.to_u32() + translation[:diff]),translation[:translated_prefix])
          else
            next unless translation[:ipv6]
            next unless translation[:original].include?(a)
            new=::IPAddress::IPv6.parse_u128((a.to_u128() + translation[:diff]),translation[:translated_prefix])
          end
        end
        if new.nil?() then
          return a.to_s
        else
          return new.to_s
        end
      end

      def translate_address(address)
        net=nil
        net=translate_network(address[:network]) unless address[:network].nil?()
        address_value=translate_address_value(address[:value])
        addr=Iof::Query::Address.new(address[:type],address_value,net,nil)
        return addr
      end

      def translate_network(network)
        value=translate_address_value(network[:value])
        net=Iof::Query::Network.new(network[:type],value,network[:netmask_length],nil)
        return net
      end

      def check_for_loop(name,rdf_data)
        result=:none
        nodes=Array.new()
        node_neighbors=Hash.new()
        seen_vlans=Array.new()
        topology=Iof::Query::VisaRDF.import(rdf_data)
        topology.each do |v|
          v[:devices].each do |d|
            nodes << d[:internal_name].intern
            node_neighbors[d[:internal_name].intern]=Array.new() unless node_neighbors.include?(d[:internal_name].intern)
            d[:interfaces].each do |i|
              next if i[:connected].nil?()
              neighbor=Hash.new()
              neighbor[:n1]=d[:internal_name].intern
              neighbor[:n1_l1]=i[:internal_name].intern
              neighbor[:n2]=i[:connected].to_s.sub(/___.*/,'').intern
              neighbor[:n2_l1]=i[:connected].intern
              neighbor[:vlans]=Array.new()
              i[:vlans].each do |vlan|
                neighbor[:vlans] << vlan[:id]
                seen_vlans << vlan[:id]
              end unless i[:vlans].nil?()
              neighbor[:vlans].uniq!
              node_neighbors[neighbor[:n2].intern].each do |n|
                next unless n[:n2_l1]==i[:internal_name].intern
                vlans=Array.new()
                vlans+=neighbor[:vlans].clone() if neighbor.include?(:vlans)
                vlans+=n[:vlans].clone() if n.include?(:vlans)
                vlans.uniq!
                vlans.compact!
                neighbor[:vlans]=vlans.clone()
                n[:vlans]=vlans.clone()
              end if node_neighbors.include?(neighbor[:n2].intern)
              node_neighbors[d[:internal_name].intern] << neighbor
            end unless d[:interfaces].nil?()
          end unless v[:devices].nil?()
        end unless topology.nil?()
        seen_vlans.uniq!
        seen_vlans.sort!
        seen_vlans.compact!

        if check_for_loop_helper(nodes,node_neighbors)
          warn "Topology #{name} has at least one physical loop"
          result=:layer1
          seen_vlans.each do |v|
            if check_for_loop_helper(nodes,node_neighbors,v)
              warn "Topology #{name} has at least one loop for vlan #{v}"
              result=:layer2
              break
            end
          end
        end
        return result
      end

      def check_for_loop_helper(nodes,node_neighbors,vlan_id=nil)
        nodes.each do |n|
          if bfs(n,node_neighbors,vlan_id)
            warn "Found loop from startnode #{n} #{(vlan_id.nil?()?("on layer 1"):("on layer 2 for vlan #{vlan_id}"))}"
            return true
          end
        end
        return false
      end

      def bfs(start_node,node_neighbors,vlan_id=nil)
        debug "BFS for start node #{start_node} #{(vlan_id.nil?()?("on layer 1"):("on layer 2 for vlan #{vlan_id}"))}"
        visited=Hash.new
        queue=Array.new()
        queue.push([start_node,nil])
        until queue.empty?() do
          node=queue.shift()
          break if node.nil?()
          if visited[node[0].intern].nil?()
            visited[node[0].intern]=1
          elsif
            warn "Double visit for #{node[0].to_s}"
#            puts "Would return true"
            return true
          end
          neighbors=node_neighbors[node[0]]
          neighbors.each do |n|
             if node[1].nil?()
              queue.push([n[:n2],n]) if (vlan_id.nil?() || (!vlan_id.nil?() && n[:vlans].include?(vlan_id)))
            else
              queue.push([n[:n2],n]) if (node[1][:n2]!=n[:n1] || node[1][:n2]!=n[:n1] || node[1][:n1_l1]!=n[:n2_l1] || node[1][:n2_l1]!=n[:n1_l1]) && (vlan_id.nil?() || (!vlan_id.nil?() && n[:vlans].include?(vlan_id)))
            end
          end unless neighbors.nil?()
        end
        return false
      rescue => e
        error "Failed to calculate path for start node #{start_node}, assuming loop found", e
        puts "Failed to calculate path for start node #{start_node}, assuming loop found", e
        return true
      end

    end
  end
end
