# encoding: utf-8
# license: gpl3p

### BEGIN LICENSE NOTICE
# This file is part of the interconnect asset ontology framework (IO)
# Copyright (C) 2010 - 2013
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
### END LICENSE NOTICE

### BEGIN AUTHOR LIST
#
### END AUTHOR LIST
module Iof
  module Data
    class Handler < Ghun::Base::Module
      def initialize(fast_writing=true,ontology_file=nil)
        super(Ghun::Log::Source::ONTOLOGY)
        @component_mutex=Monitor.new()
        @connection_mutex=Monitor.new()
        @ontology_mutex=Monitor.new()
        @components=Hash.new()
        @neighbors=Array.new()
        @connections=Array.new()
        @mapper=Iof::Data::Mapper.instance()
        @ontology=Iof::Ontology::JenaGlue.new(false,true,true,ontology_file)
        @fast_writing=fast_writing
        if @fast_writing then
          @ontology.start_for_writing()
        else
          @ontology.start()
        end
        #TODO extend on interfaces
        @changed_names=Hash.new()
        @known_connections=Hash.new()
        @static=Hash.new()
        @static[:vlan]=Hash.new
        @static[:mac]=Hash.new
        @static[:ipv4_address]=Hash.new
        @static[:ipv6_address]=Hash.new
        @static[:ipv4_network]=Hash.new
        @static[:ipv6_network]=Hash.new
        @static[:icmpv4]=Hash.new
        @static[:icmpv6]=Hash.new
        @static[:udp]=Hash.new
        @static[:tcp]=Hash.new
        @static_mutex=Mutex.new()
        @group=Hash.new()
        @group[:vsa]=Hash.new
        @group_mutex=Mutex.new()
        @member=Hash.new()
        @member[:vsa]=Hash.new()
        @member_mutex=Mutex.new()
      end
      attr_reader :ontology, :mapper

      def write_changes(restart_ontology=true)
        store_static()
        store_components()
        store_neighbors()
        store_connections()
        store_groups()
        @ontology_mutex.synchronize {
          @ontology.finish()
          if @fast_writing then
            @ontology.start_for_writing()
          else
            @ontology.start()
          end if restart_ontology
        }
      end

      def get_static(type,*args)
        id=nil
        name=nil
        klass=nil
        values=Hash.new()
        case type
        when :mac then
          klass=Iof::Data::NO::MacAddress
          id=args[0].to_s.intern
          name="__mac_#{args[0].to_s}__"
          values[:no_mac_address_value]=args[0].to_s
        when :vlan then
          klass=Iof::Data::NO::EthernetVlan
          id=args[0].to_s.intern
          name="__vlan_#{args[0].to_s}__"
          values[:no_vlan_id]=args[0].to_s
        when :icmpv4 then
          klass=Iof::Data::NO::ICMPv4Type
          id=args[0].to_s.intern
          name="__icmpv4_#{args[0].to_s}__"
          values[:no_icmpv4_type_value]=args[0].to_s
        when :icmpv6 then
          klass=Iof::Data::NO::ICMPv6Type
          id=args[0].to_s.intern
          name="__icmpv6_#{args[0].to_s}__"
          values[:no_icmpv6_type_value]=args[0].to_s
        when :udp then
          klass=Iof::Data::NO::UDPPort
          id=args[0].to_s.intern
          name="__udp_#{args[0].to_s}__"
          values[:no_udp_port_value]=args[0].to_s
        when :tcp then
          klass=Iof::Data::NO::TCPPort
          id=args[0].to_s.intern
          name="__tcp_#{args[0].to_s}__"
          values[:no_tcp_port_value]=args[0].to_s
        when :ipv4_address then
          klass=Iof::Data::NO::IPv4Address
          values[:no_ipv4_address_value]=args[0].to_s
          if args.size==1 then
            id=args[0].to_s.intern
            name="__ipv4_address_#{args[0].to_s}__"
          else
            netmask=Iof::Data.ipv4_parse_netmask(args[1])
            length=Iof::Data.ipv4_netmask_to_length(netmask)
            id="#{args[0].to_s}/#{length}".intern
            name="__ipv4_address_#{args[0].to_s}/#{length}__"
            values[:no_ipv4_netmask_value]=netmask
            values[:no_ipv4_netmask_length]=length
          end
        when :ipv4_network then
          klass=Iof::Data::NO::IPv4Network
          values[:no_ipv4_address_value]=args[0].to_s
          if args.size==1 then
            id=args[0].to_s.intern
            name="__ipv4_network_#{args[0].to_s}__"
          else
            netmask=Iof::Data.ipv4_parse_netmask(args[1])
            length=Iof::Data.ipv4_netmask_to_length(netmask)
            id="#{args[0].to_s}/#{length}".intern
            name="__ipv4_network_#{args[0].to_s}/#{length}__"
            values[:no_ipv4_netmask_value]=netmask
            values[:no_ipv4_netmask_length]=length
          end
        when :ipv6_address then
          klass=Iof::Data::NO::IPv6Address
          values[:no_ipv6_address_value]=args[0].to_s
          if args.size==1 then
            id=args[0].to_s.intern
            name="__ipv6_address_#{args[0].to_s}__"
          else
            id="#{args[0].to_s}/#{args[1].to_s}".intern
            name="__ipv6_address_#{args[0].to_s}/#{args[1]}__"
            values[:no_ipv6_netmask_length]=args[1].to_s
          end
        when :ipv6_network then
          klass=Iof::Data::NO::IPv6Network
          values[:no_ipv6_address_value]=args[0].to_s
          if args.size==1 then
            id=args[0].to_s.intern
            name="__ipv6_network_#{args[0].to_s}__"
          else
            id="#{args[0].to_s}/#{args[1].to_s}".intern
            name="__ipv6_network_#{args[0].to_s}/#{args[1]}__"
            values[:no_ipv6_netmask_length]=args[1].to_s
          end
        end
        static=nil
        @static_mutex.synchronize {
          if @static[type].include?(id) then
            static=@static[type][id]
          else
            static=klass.new()
            static.name=name
            values.each do |key,value|
              static[key]=value
            end
            @static[type][id]=static
          end
        }
        return static
      rescue => e
        error "Failed to get individual for type #{type.to_s} and arguments #{args.inspect}", e
        return nil
      end

      def get_static_name(type,*args)
        static=get_static(type,*args)
        if static then
          return static.name
        else
          return nil
        end
      end

      def get_mac_address_name(mac)
        get_static_name(:mac,Iof::Data.mac_plain(mac))
      end

      def get_vlan_name(vlan_id)
        get_static_name(:vlan,vlan_id.to_s)
      end

      def get_udp_port_name(port)
        get_static_name(:udp,port.to_s)
      end

      def get_tcp_port_name(port)
        get_static_name(:tcp,port.to_s)
      end

      def get_icmpv4_type_name(type)
        get_static_name(:icmpv4,type.to_s)
      end

      def get_icmpv6_type_name(type)
        get_static_name(:icmpv6,type.to_s)
      end

      def get_ipv4_address_name(address,netmask=nil)
        if netmask.nil?() then
          get_static_name(:ipv4_address,address)
        else
          get_static_name(:ipv4_address,address,netmask)
        end
      end

      def get_ipv4_network_name(address,netmask)
        get_static_name(:ipv4_network,address,netmask)
      end

      def get_ipv6_address_name(address,netmask=nil)
        if netmask.nil?() then
          get_static_name(:ipv6_address,Iof::Data.ipv6_address_uncompressed(address))
        else
          get_static_name(:ipv6_address,Iof::Data.ipv6_address_uncompressed(address),netmask)
        end
      end

      def get_ipv6_network_name(address,netmask)
        get_static_name(:ipv6_network,Iof::Data.ipv6_address_uncompressed(address),netmask)
      end

      def get_group_name(type,*args)
        group=get_group(type,*args)
        if group then
          return group.name
        else
          return nil
        end
      end

      def get_group(type,*args)
        id=nil
        name=nil
        klass=nil
        values=Hash.new()
        case type
        when :vsa then
          klass=Iof::Data::NO::VSA
          id=args[0].to_s.intern
          name="__vsa_#{args[0].to_s}__"
        end
        group=nil
        @group_mutex.synchronize {
          if @group[type].include?(id) then
            group=@group[type][id]
          else
            group=klass.new()
            group.name=name
            values.each do |key,value|
              group[key]=value
            end
            @group[type][id]=group
            @member_mutex.synchronize {
              @member[type][id]=Hash.new()
            }
          end
        }
        return group
      rescue => e
        error "Failed to get individual for type #{type.to_s} and arguments #{args.inspect}", e
        return nil
      end

      def set_group_member(type,group_name,member_name)
        @member_mutex.synchronize {
          unless @member[type][group_name.to_s.intern].include?(member_name.intern) then
            membership=Iof::Data::NO::HasVSAMember.new()
            membership.subject=get_group_name(type, group_name)
            membership.object=member_name
            @member[type][group_name.to_s.intern][member_name.intern]=membership
          end
        }
      end

      def set_component(component)
        @component_mutex.synchronize {
          if @components[component.name.intern].nil? then
            @components[component.name.intern]=component
          else
            warn "Overwriting unsafed state for component '#{component.name}'"
            @components[component.name.intern]=component
          end
        }
      end

      def set_neighbor(neighbor)
        @component_mutex.synchronize {
          @neighbors << neighbor
        }
      end

      def set_connection(connection)
        @connection_mutex.synchronize {
          @connections << connection
        }
      end

    private

      def store_static
        @static_mutex.synchronize {
          @ontology_mutex.synchronize {
            @static.each do |type,data|
              debug "Storing #{type.to_s}"
              data.each do |id,d|
                next if d.nil?()
                d.store(@ontology,nil,:no)
              end
            end
          }
        }
      end

      def store_components
        debug "Storing components"
        components=nil
        @component_mutex.synchronize {
          components=@components
          @components=Hash.new()
        }
        error "Got no components" if components.empty?()
        components.each do |name,e|
          if e.nil?() then
            error "Got empty component"
            next
          end
          @ontology_mutex.synchronize {
            name=e.name
            stored=e.store(@ontology,nil,:no)
            unless name==stored[:name] then
              @changed_names[name.intern]=stored
              debug "Name for component '#{name}' changed to '#{stored}'"
            end
            #drop all connections, store_connections will restore still existing ones
            drop_connections(stored,:no)
          }
        end
      end

      def store_neighbors
        debug "Storing neighbors"
        neighbors=nil
        @component_mutex.synchronize {
          neighbors=@neighbors
          @neighbors=Hash.new()
        }
        error "Got no neighbors" if neighbors.empty?()
        neighbors.each do |e|
          if e.nil?() then
            error "Got empty neighbor"
            next
          end
          @ontology_mutex.synchronize {
            name=e.name
            stored=e.store(@ontology,nil,:no,nil,nil,[:no_identifier],true)
            unless name==stored[:name] then
              @changed_names[name.intern]=stored
              debug "Name for neighbor '#{name}' changed to '#{stored}'"
            end
            #drop all connections, store_connections will restore still existing ones
            drop_connections(stored,:no)
          }
        end
      end

      def drop_connections(component,ns)
        url=Iof::Data.namespace_to_url(ns)
        result=@ontology.query("SELECT ?l1 WHERE { <#{url}#{component[:name]}> no:hasLayer1Interface ?l1 }")
        result.each do |line|
          @ontology.drop_object_property(line[:l1], "isConnectedTo", line[:__ns_l1], :no)
        end
      end

      def store_connections()
        debug "Storing connections"
        connections=nil
        @connection_mutex.synchronize {
          connections=@connections
          @connections=Array.new()
        }
        @ontology_mutex.synchronize {
          connections.each do |c|
            begin
              if c.is_a?(Iof::Data::NO::IsConnectedTo) then
                debug "Storing property #{c.class.to_s} with subject '#{c.subject}' and object '#{c.object}'"
                subject_component=c.subject[:name].sub(/___.*/,'')
                object_component=c.object[:name].sub(/___.*/,'')
                if @changed_names.include?(subject_component.intern) then
                  new_subject=c.subject[:name].sub(/^#{subject_component}___/,"#{@changed_names[subject_component.intern][:name]}___")
                  debug "Changed subject from '#{c.subject[:name]}' to '#{new_subject}'"
                  c.subject=new_subject
                end
                if @changed_names.include?(object_component.intern) then
                  new_object=c.object[:name].sub(/^#{object_component}___/,"#{@changed_names[object_component.intern][:name]}___")
                  debug "Changed object from '#{c.object[:name]}' to '#{new_object}'"
                  c.object=new_object
                end
                if (@known_connections.include?(c.subject[:name].intern) && @known_connections[c.subject[:name].intern].include?(c.object[:name].intern)) || (@known_connections.include?(c.object[:name].intern) && @known_connections[c.object[:name].intern].include?(c.subject[:name].intern)) then
                  debug "Connection already stored"
                else
                  c.store(@ontology,:no)
                  @known_connections[c.subject[:name].intern]=Array.new() unless @known_connections.include?(c.subject[:name].intern)
                  @known_connections[c.subject[:name].intern] << c.object[:name].intern
                  @known_connections[c.object[:name].intern]=Array.new() unless @known_connections.include?(c.object[:name].intern)
                  @known_connections[c.object[:name].intern] << c.subject[:name].intern
                end
              else
                debug "Not storing non connection property #{c.class.to_s} with subject '#{c.subject}' and object '#{c.object}'"
              end
            rescue => e
              error "Failed to store property #{c.class.to_s} with subject '#{c.subject}' and object '#{c.object}'", e
            end
          end
        }
      end

      def store_groups()
        debug "Storing groups"
        @group_mutex.synchronize {
          @ontology_mutex.synchronize {
            @group.each do |type,data|
              debug "Storing #{type.to_s}"
              data.each do |id,d|
                next if d.nil?()
                d.store(@ontology,nil,:no)
              end
            end
          }
        }

        debug "Storing group members"
        @member_mutex.synchronize {
          @ontology_mutex.synchronize {
            @member.each do |type,data|
              debug "Storing #{type.to_s} members"
              data.each do |id,d|
                d.each do |m_id,m|
                  next if m.nil?()
                  m.store(@ontology,:no)
                end unless d.nil?()
              end
            end
          }
        }
      end
    end
  end
end
